from pathlib import PurePath, PurePosixPath
from typing import IO, TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union


from .workspace import Workspace
from .job import Job, ExternalJob
from .dataset import Dataset, DEFAULT_FORMAT
from .row import R
from .spec import MongoController, Datafield, Datatype, ProjectDocument

if TYPE_CHECKING:
    from numpy.typing import NDArray  # type: ignore
    from .tools import CryoSPARC


class Project(MongoController[ProjectDocument]):
    """
    Accessor instance for CryoSPARC projects with ability to add workspaces, jobs
    and upload/download project files. Should be instantiated through
    `CryoSPARC.find_project`_.

    .. _CryoSPARC.find_project:
        tools.html#cryosparc.tools.CryoSPARC.find_project
    """

    def __init__(self, cs: "CryoSPARC", uid: str) -> None:
        self.cs = cs
        self.uid = uid

    def refresh(self):
        """
        Reload this project from the CryoSPARC database.

        Returns:
            Project: self
        """
        self._doc = self.cs.cli.get_project(self.uid)  # type: ignore
        return self

    def dir(self) -> PurePosixPath:
        """
        Get the path to the project directory.

        Returns:
            Path: project directory Pure Path instance
        """
        path: str = self.cs.cli.get_project_dir_abs(self.uid)  # type: ignore
        return PurePosixPath(path)

    def find_workspace(self, workspace_uid) -> Workspace:
        """
        Get a workspace accessor instance for the workspace in this project
        with the given UID. Fails with an error if workspace does not exist.

        Args:
            workspace_uid (str): Workspace unique ID, e.g., "W1"

        Returns:
            Workspace: accessor instance
        """
        workspace = Workspace(self.cs, self.uid, workspace_uid)
        return workspace.refresh()

    def find_job(self, job_uid: str) -> Job:
        """
        Get a job accessor instance for the job in this project with the given
        UID. Fails with an error if job does not exist.

        Args:
            job_uid (str): Job unique ID, e.g., "J42"

        Returns:
            Job: accessor instance
        """
        job = Job(self.cs, self.uid, job_uid)
        job.refresh()
        return job

    def find_external_job(self, job_uid: str) -> ExternalJob:
        """
        Get the External job accessor instance for an External job in this
        project with the given UID. Fails if the job does not exist or is not an
        external job.

        Args:
            job_uid (str): Job unique ID, e.g,. "J42"

        Raises:
            TypeError: If job is not an external job

        Returns:
            ExternalJob: accessor instance
        """
        return self.cs.find_external_job(self.uid, job_uid)

    def create_workspace(self, title: str, desc: Optional[str] = None) -> Workspace:
        """
        Create a new empty workspace in this project. At least a title must be
        provided.

        Args:
            title (str): Title of new workspace
            desc (str, optional): Markdown text description. Defaults to None.

        Returns:
            Workspace: created workspace instance
        """
        return self.cs.create_workspace(self.uid, title, desc)

    def create_job(
        self,
        workspace_uid: str,
        type: str,
        connections: Dict[str, Tuple[str, str]] = {},
        params: Dict[str, Any] = {},
        title: Optional[str] = None,
        desc: Optional[str] = None,
    ) -> Job:
        """
        Create a new job with the given type. Use `CryoSPARC.get_job_sections`_
        to query available job types on the connected CryoSPARC instance.

        Args:
            project_uid (str): Project UID to create job in, e.g., "P3"
            workspace_uid (str): Workspace UID to create job in, e.g., "W1"
            type (str): Job type identifier, e.g., "homo_abinit"
            connections (dict[str, tuple[str, str]]): Initial input connections.
                Each key is an input name and each value is a (job uid, output
                name) tuple. Defaults to {}
            params (dict[str, any], optional): Specify parameter values.
                Defaults to {}.
            title (str, optional): Job title. Defaults to None.
            desc (str, optional): Job markdown description. Defaults to None.

        Returns:
            Job: created job instance. Raises error if job cannot be created.

        Examples:

            Create an Import Movies job.

            >>> from cryosparc.tools import CryoSPARC
            >>> cs = CryoSPARC()
            >>> project = cs.find_project("P3")
            >>> import_job = project.create_job("W1", "import_movies")
            >>> import_job.set_param("blob_paths", "/bulk/data/t20s/*.tif")
            True

            Create a 3-class ab-initio job connected to existing particles.

            >>> abinit_job = project.create_job("W1", "homo_abinit"
            ...     connections={"particles": ("J20", "particles_selected")}
            ...     params={"abinit_K": 3}
            ... )

        .. _CryoSPARC.get_job_sections:
            tools.html#cryosparc.tools.CryoSPARC.get_job_sections
        """
        return self.cs.create_job(
            self.uid, workspace_uid, type, connections=connections, params=params, title=title, desc=desc
        )

    def create_external_job(
        self,
        workspace_uid: str,
        title: Optional[str] = None,
        desc: Optional[str] = None,
    ) -> ExternalJob:
        """
        Add a new External job to this project to save generated outputs to.

        Args:
            workspace_uid (str): Workspace UID to create job in, e.g., "W3".
            title (str, optional): Title for external job (recommended).
                Defaults to None.
            desc (str, optional): Markdown description for external job.
                Defaults to None.

        Returns:
            ExternalJob: created external job instance
        """
        job_uid: str = self.cs.vis.create_external_job(  # type: ignore
            project_uid=self.uid, workspace_uid=workspace_uid, user=self.cs.user_id, title=title, desc=desc
        )
        return self.find_external_job(job_uid)

    def save_external_result(
        self,
        workspace_uid: Optional[str],
        dataset: Dataset[R],
        type: Datatype,
        name: Optional[str] = None,
        slots: Optional[List[Union[str, Datafield]]] = None,
        passthrough: Optional[Tuple[str, str]] = None,
        title: Optional[str] = None,
        desc: Optional[str] = None,
    ) -> str:
        """
        Save the given result dataset to the project. Specify at least the
        dataset to save and the type of data.

        If ``workspace_uid`` or ``passthrough`` input is not specified or is
        None, saves the result to the project's newest workspace. If
        ``passthrough`` is specified but ``workspace_uid`` is not, saves to the
        passthrough job's newest workspace.

        Returns UID of the External job where the results were saved.

        Examples:

            Save all particle data

            >>> particles = Dataset()
            >>> project.save_external_result("W1", particles, 'particle')
            "J43"

            Save new particle locations that inherit passthrough slots from a
            parent job

            >>> particles = Dataset()
            >>> project.save_external_result(
            ...     workspace_uid='W1',
            ...     dataset=particles,
            ...     type='particle',
            ...     name='particles',
            ...     slots=['location'],
            ...     passthrough=('J42', 'selected_particles'),
            ...     title='Re-centered particles'
            ... )
            "J44"

            Save a result with multiple slots of the same type.

            >>> project.save_external_result(
            ...     workspace_uid="P1",
            ...     dataset=particles,
            ...     type="particle",
            ...     name="particle_alignments",
            ...     slots=[
            ...         {"dtype": "alignments3D", "prefix": "alignments_class_0", "required": True},
            ...         {"dtype": "alignments3D", "prefix": "alignments_class_1", "required": True},
            ...         {"dtype": "alignments3D", "prefix": "alignments_class_2", "required": True},
            ...     ]
            ... )
            "J45"

        Args:
            workspace_uid (str | None): Workspace UID to save results into.
                Specify ``None`` to auto-select a workspace.
            dataset (Dataset): Result dataset.
            type (Datatype): Type of output dataset.
            name (str, optional): Name of output on created External job. Same
                as type if unspecified. Defaults to None.
            slots (list[str | Datafield], optional): List of slots expected to
                be created for this output such as ``location`` or ``blob``. Do
                not specify any slots that were passed through from an input
                unless those slots are modified in the output. Defaults to None.
            passthrough (tuple[str, str], optional): Indicates that this output
                inherits slots from the specified output. e.g., ``("J1",
                "particles")``. Defaults to None.
            title (str, optional): Human-readable title for this output.
                Defaults to None.
            desc (str, optional): Markdown description for this output. Defaults
                to None.

        Returns:
            str: UID of created job where this output was saved
        """
        return self.cs.save_external_result(
            self.uid,
            workspace_uid,
            dataset=dataset,
            type=type,
            name=name,
            slots=slots,
            passthrough=passthrough,
            title=title,
            desc=desc,
        )

    def list_files(self, prefix: Union[str, PurePosixPath] = "", recursive: bool = False) -> List[str]:
        """
        Get a list of files inside the project directory.

        Args:
            prefix (str | Path, optional): Subdirectory inside project to list.
                Defaults to "".
            recursive (bool, optional): If True, lists files recursively.
                Defaults to False.

        Returns:
            list[str]: List of file paths relative to the project directory.
        """
        return self.cs.list_files(self.uid, prefix=prefix, recursive=recursive)

    def download(self, path_rel: Union[str, PurePosixPath]):
        """
        Open a file in the current project for reading. Use to get files from a
        remote CryoSPARC instance where the project directory is not available
        on the client file system.

        Args:
            path (str | Path): Relative path of file in project directory.

        Yields:
            HTTPResponse: Use a context manager to read the file from the
            request body
        """
        return self.cs.download(self.uid, path_rel)

    def download_file(self, path_rel: Union[str, PurePosixPath], target: Union[str, PurePath, IO[bytes]]):
        """
        Download a file from the project directory to the given writeable target.

        Args:
            path_rel (str | Path): Relative path of file in project directory.
            target (str | Path | IO): Local file path, directory path or writeable
                file handle to write response data.

        Returns:
            Path | IO: resulting target path or file handle.
        """
        return self.cs.download_file(self.uid, path_rel, target)

    def download_dataset(self, path_rel: Union[str, PurePosixPath]):
        """
        Download a .cs dataset file frmo the given relative path in the project
        directory.

        Args:
            path_rel (str | Path): Realtive path to .cs file in project
            directory.

        Returns:
            Dataset: Loaded dataset instance
        """
        return self.cs.download_dataset(self.uid, path_rel)

    def download_mrc(self, path_rel: Union[str, PurePosixPath]):
        """
        Download a .mrc file from the given relative path in the project
        directory.

        Args:
            path_rel (str | Path): Relative path to .mrc file in project
                directory.

        Returns:
            tuple[Header, NDArray]: MRC file header and data as a numpy array
        """
        return self.cs.download_mrc(self.uid, path_rel)

    def upload(self, target_path_rel: Union[str, PurePosixPath], source: Union[str, bytes, PurePath, IO]):
        """
        Upload the given file to the project directory at the given relative
        path.

        Args:
            target_path_rel (str | Path): Relative target path in project
                directory.
            source (str | bytes | Path | IO): Local path or file handle to
                upload. May also specified as raw bytes.
        """
        return self.cs.upload(self.uid, target_path_rel, source)

    def upload_dataset(self, target_path_rel: Union[str, PurePosixPath], dset: Dataset, format: int = DEFAULT_FORMAT):
        """
        Upload a dataset as a CS file into the project directory.

        Args:
            target_path_rel (str | Path): relative path to save dataset in
                project directory. Should have a ``.cs`` extension.
            dset (Dataset): dataset to save.
            format (int): format to save in from ``cryosparc.dataset.*_FORMAT``,
                defaults to NUMPY_FORMAT)

        """
        return self.cs.upload_dataset(self.uid, target_path_rel, dset, format=format)

    def upload_mrc(self, target_path_rel: Union[str, PurePosixPath], data: "NDArray", psize: float):
        """
        Upload a numpy 2D or 3D array to the job directory as an MRC file.

        Args:
            target_path_rel (str | Path): filename or relative path. Should have
                ``.mrc`` extension.
            data (NDArray): Numpy array with MRC file data.
            psize (float): Pixel size to include in MRC header.
        """
        return self.cs.upload_mrc(self.uid, target_path_rel, data, psize)

    def mkdir(
        self,
        target_path_rel: Union[str, PurePosixPath],
        parents: bool = False,
        exist_ok: bool = False,
    ):
        """
        Create a directory in the given project.

        Args:
            target_path_rel (str | Path): Relative path to create inside project
                directory.
            parents (bool, optional): If True, any missing parents are created
                as needed. Defaults to False.
            exist_ok (bool, optional): If True, does not raise an error for
                existing directories. Still raises if the target path is not a
                directory. Defaults to False.
        """
        self.cs.mkdir(
            project_uid=self.uid,
            target_path_rel=target_path_rel,
            parents=parents,
            exist_ok=exist_ok,
        )

    def cp(self, source_path_rel: Union[str, PurePosixPath], target_path_rel: Union[str, PurePosixPath]):
        """
        Copy a file or folder within a project to another location within that
        same project. Note that argument order is reversed from
        equivalent ``cp`` command.

        Args:
            source_path_rel (str | Path): Relative path in project of source
                file or folder to copy.
            target_path_rel (str | Path): Relative path in project to copy to.
        """
        self.cs.cp(
            project_uid=self.uid,
            source_path_rel=source_path_rel,
            target_path_rel=target_path_rel,
        )

    def symlink(self, source_path_rel: Union[str, PurePosixPath], target_path_rel: Union[str, PurePosixPath]):
        """
        Create a symbolic link in the given project. May only create links for
        files within the project. Note that argument order is reversed from
        ``ln -s``.

        Args:
            project_uid (str): Target project UID, e.g., "P3".
            target_path_rel (str | Path): Relative path in project to new
                symlink.
            source_path_rel (str | Path): Relative path in project to file from
                which to create symlink.
        """
        self.cs.symlink(
            project_uid=self.uid,
            source_path_rel=source_path_rel,
            target_path_rel=target_path_rel,
        )



from libc.stdint cimport uint64_t, uint32_t
ctypedef uint64_t Dset

cdef extern from "cryosparc-tools/dataset.h":

    Dset dset_new() nogil
    Dset dset_copy(Dset dset) nogil
    Dset dset_innerjoin(const char *key, Dset dset_r, Dset dset_s) nogil
    void dset_del(Dset dset) nogil

    uint64_t dset_totalsz(Dset dset) nogil
    uint32_t dset_ncol(Dset dset) nogil
    uint64_t dset_nrow(Dset dset) nogil
    const char *dset_key(Dset dset, uint64_t index) nogil
    int dset_type(Dset dset, const char *colkey) nogil
    void *dset_get(Dset dset, const char *colkey) nogil
    uint64_t dset_getsz(Dset dset, const char *colkey) nogil
    bint dset_setstr(Dset dset, const char *colkey, uint64_t index, const char *value) nogil
    const char *dset_getstr(Dset dset, const char *colkey, uint64_t index) nogil
    uint32_t dset_getshp(Dset dset, const char *colkey) nogil

    bint dset_addrows(Dset dset, uint32_t num) nogil
    bint dset_addcol_scalar(Dset dset, const char *key, int type) nogil
    bint dset_addcol_array(Dset dset, const char *key, int type, int shape0, int shape1, int shape2) nogil
    bint dset_changecol(Dset dset, const char *key, int type) nogil
    bint dset_defrag(Dset dset, bint realloc_smaller) nogil

    void dset_dumptxt(Dset dset) nogil



__version__ = "4.1.2"


def get_include():
    """
    Get the include directory for the ``<cryosparc-tools/dataset.h>`` header
    file to access CryoSPARC dataset handles from other languages.

    Returns:
        str: Include path for C compilers.
    """
    from pathlib import Path

    return str(Path(__file__).parent / "include")



from typing import Any, Dict, Generic, Iterable, List, Mapping, Optional, Tuple, TypeVar

import numpy as n

from .column import Column
from .util import default_rng, random_integers


class Row(Mapping):
    """
    Provides row-by-row access to a dataset. Do not initialize directly. See
    dataset module.
    """

    __slots__ = ("idx", "cols", "__dict__")  # Specifying this speeds up allocation of many rows
    idx: int
    cols: Dict[str, Column]

    def __init__(self, cols: Dict[str, Column], idx: int):
        self.idx = idx
        self.cols = cols
        # note - don't keep around a ref to cols because then when col._data
        # changes (e.g., a field is added to the dataset) the already existing
        # items will be referring to the old dataset._data!

    def __array__(self, dtype=None):
        # Prevent numpy from attempting to turn row to array of strings
        a = n.ndarray(shape=(), dtype=n.dtype(object))
        a[()] = self
        return a

    def __len__(self):
        return len(self.cols)

    def __getitem__(self, key: str):
        return self.cols[key][self.idx]

    def __setitem__(self, key: str, value):
        self.cols[key][self.idx] = value

    def __contains__(self, key: object):
        return key in self.cols

    def __iter__(self):
        return iter(self.cols)

    def item(self, key, default=None):
        return self.cols[key].item(self.idx) if key in self else default

    def to_list(self, exclude_uid=False):
        """
        Convert into a list of native python types in the same order as the
        declared fields.
        """
        return [self.cols[key].item(self.idx) for key in self.cols if not exclude_uid or key != "uid"]

    def to_dict(self):
        return {key: self.cols[key].item(self.idx) for key in self.cols}

    def from_dict(self, d):
        for k in self.cols:
            self[k] = d[k]

    def __repr__(self) -> str:
        result = f"{type(self).__name__}(["
        for k in self:
            val = self[k]
            result += f"\n    ('{k}', {val if isinstance(val, n.ndarray) else repr(val)}),"
        return result + "\n])"

    def _ipython_key_completions_(self):
        return list(self.cols)


R = TypeVar("R", bound=Row)
"""
Type variable for a ``Row`` subclass.
"""


class Spool(List[R], Generic[R]):
    """
    List-like dataset row accessor class with support for splitting and
    randomizing based on row fields

    Args:
        items (Iterable[R]): List of rows
        rng (Generator, optional): Numpy random number generator. Uses
            ``numpy.random.default_rng()`` if not specified. Defaults to None.
    """

    DEFAULT_RNG = default_rng()

    @classmethod
    def set_default_random(cls, rng: "n.random.Generator"):
        """
        Reset the default random number generator for all Spools

        Args:
            rng (Generator): Numpy random generator.
        """
        cls.DEFAULT_RNG = rng

    def __init__(self, items: Iterable[R], rng: "Optional[n.random.Generator]" = None):
        super().__init__(items)
        self.indices = None
        self.random = rng or self.DEFAULT_RNG

    def set_random(self, rng: "n.random.Generator"):
        """
        Reset the random number generator for this Spool.

        Args:
            rng (Generator): Numpy random generator.
        """
        self.random = rng

    # -------------------------------------------------- Spooling and Splitting
    def split(self, num: int, random: bool = True, prefix: Optional[str] = None):
        """
        Return two Spools with the elements of this spool. The first Spool
        contains ``num`` elements. The second contains ``len(self) - num``
        elements.

        Args:
            num (int): Number of elements in first split.
            random (bool, optional): If True, add elements to each split in
                random order. Defaults to True.
            prefix (str, optional): If specified, set each ``{prefix}/split``
                field to ``0`` in the first split and ``1`` in the second split.
                Defaults to None.

        Returns:
            tuple[Spool, Spool]: the two split lists
        """
        if random:
            idxs = self.random.permutation(len(self))
        else:
            idxs = n.arange(len(self))
        d1 = Spool(items=(self[i] for i in idxs[:num]), rng=self.random)
        d2 = Spool(items=(self[i] for i in idxs[num:]), rng=self.random)
        if prefix is not None:
            field = prefix + "/split"
            for img in d1:
                img[field] = 0
            for img in d2:
                img[field] = 1
        return d1, d2

    def split_half_in_order(self, prefix: str, random: bool = True):
        """
        Split into two spools of approximately equal size. Elements are added
        to each split in stable order.

        Args:
            prefix (str): Set each element's ``{prefix}/split`` field to ``0``
                or ``1`` depending on which split it's added to.
            random (bool, optional): If True, randomly assign each element to a
                split. Defaults to True.

        Returns:
            tuple[Spool, Spool]: the two split lists
        """
        if random:
            splitvals = random_integers(self.random, 2, size=len(self))
        else:
            splitvals = n.arange(len(self)) % 2
        for idx, p in enumerate(self):
            p[prefix + "/split"] = splitvals[idx]
        d1 = Spool(items=(p for p in self if p[prefix + "/split"] == 0), rng=self.random)
        d2 = Spool(items=(p for p in self if p[prefix + "/split"] == 1), rng=self.random)
        return d1, d2

    def split_into_quarter(self, num: int):
        """
        Randomly assign the elements of this Spool to two new Spools. The first
        Spool contains ``num`` elements. The second contains ``len(self) - num``
        elements.

        Args:
            num (int): Number of elements in the first split.

        Returns:
            tuple[Spool, Spool]: the two split lists
        """
        return self.split(num, random=True)

    def split_by_splits(self, prefix: str = "alignments"):
        """
        Return two Spools divided by the value of the ``{prefix}/split`` field.

        Args:
            prefix (str, optional): Field to prefix to use to determine which
                element goes into which split list. Defaults to "alignments".

        Returns:
            tuple[Spool, Spool]: the two split lists
        """
        idxs = n.arange(len(self))
        field = "split"
        if prefix is not None:
            field = prefix + "/split"
        d1 = Spool(items=(self[i] for i in idxs if self[i][field] == 0), rng=self.random)
        d2 = Spool(items=(self[i] for i in idxs if self[i][field] == 1), rng=self.random)
        return d1, d2

    def split_from_field(self, field: str, vals: Tuple[Any, Any] = (0, 1)):
        """
        Split into two lists based on the given possible vals of the given
        field.

        Args:
            field (str): Split field to test for split value.
            vals (tuple[any, any], optional): Possible split field values.
                Defaults to (0, 1).

        Returns:
            tuple[Spool, Spool]: the two split lists
        """
        d1 = Spool((img for img in self if img[field] == vals[0]), rng=self.random)
        d2 = Spool((img for img in self if img[field] == vals[1]), rng=self.random)
        return d1, d2

    def split_by(self, field: str) -> Dict[Any, List[R]]:
        """
        Split into a dictionary of lists, where each key is a possible value
        for the given field and each value is a list if elements that have
        that value.

        Args:
            field (str): dataset field to split on

        Returns:
            dict[any, list[R]]: dict of split element lists
        """
        items: Dict[Any, List[R]] = {}
        for item in self:
            val = item[field]
            curr = items.get(val, [])
            curr.append(item)
            items[val] = curr
        return items

    def get_random_subset(self, num: int):
        """
        Randomly selected subset of the given size, without replacement.

        Args:
            num (int): number of elements to select from the Spool

        Returns:
            list[R]: selected elements
        """
        assert num <= len(self), "Not Enough Images!"
        idxs = self.random.choice(len(self), size=num, replace=False)
        return [self[i] for i in idxs]

    def setup_spooling(self, random=True):
        """
        Determine the iteration order for the ``spool()`` method.

        Args:
            random (bool, optional): Randomize spooling order. Defaults to True.
        """
        if random:
            self.indices = self.random.permutation(len(self))
        else:
            self.indices = n.arange(len(self))
        self.spool_index = 0

    def spool(self, num: int, peek: bool = False):
        """
        Get a list consisting of num randomly-selected elements. Advance the
        spool. If peek is true, don't advance the spool, just return the first
        num elements.

        Args:
            num (int): Number of elements to get from the spool.
            peek (bool, optional): If True, does not advance internal spool
                index. Will return the same elements on the next ``spool()``
                call. Defaults to False.

        Returns:
            list[R]: list of selected spool elements
        """
        if self.indices is None:
            self.setup_spooling()
        assert self.indices is not None
        if num >= len(self):  # asking for too many
            return [self[i] for i in range(len(self))]  # just return self, no random order.
        current_indices = n.arange(self.spool_index, self.spool_index + num) % len(self.indices)
        if not peek:
            self.spool_index = (self.spool_index + num) % len(self.indices)
        return [self[self.indices[i]] for i in current_indices]

    def make_batches(self, num: int = 200):
        """
        Get a list of lists, each one a consecutive list of ``num`` images.

        Args:
            num (int, optional): Size of each batch. Defaults to 200.

        Returns:
            list[list[R]]: Split batches
        """
        return [self[idx : idx + num] for idx in n.arange(0, len(self), num)]

    def __str__(self):
        s = f"{type(self).__name__} object with {len(self)} items."
        return s



"""
Defines the Job and External job classes for accessing CryoSPARC jobs.
"""
from contextlib import contextmanager
from io import BytesIO
import json
from pathlib import PurePath, PurePosixPath
from time import sleep, time
from typing import IO, TYPE_CHECKING, Any, Iterable, List, Optional, Pattern, Union, overload
from typing_extensions import Literal

from .command import make_json_request, make_request
from .dataset import Dataset, DEFAULT_FORMAT
from .spec import (
    ASSET_CONTENT_TYPES,
    IMAGE_CONTENT_TYPES,
    TEXT_CONTENT_TYPES,
    AssetDetails,
    AssetFormat,
    MongoController,
    ImageFormat,
    JobStatus,
    TextFormat,
    EventLogAsset,
    Datatype,
    Datafield,
    JobDocument,
)
from .util import bopen, first


if TYPE_CHECKING:
    from numpy.typing import NDArray, ArrayLike
    from .tools import CryoSPARC


class Job(MongoController[JobDocument]):
    """
    Accessor class to a job in CryoSPARC with ability to load inputs and
    outputs, add to job log, download job files. Should be instantiated
    through `CryoSPARC.find_job`_ or `Project.find_job`_.

    Examples:

        Find an existing job.

        >>> cs = CryoSPARC()
        >>> job = cs.find_job("P3", "J42")
        >>> job.status
        "building"

        Queue a job.

        >>> job.queue("worker_lane")
        >>> job.status
        "queued"

        Create a 3-class ab-initio job connected to existing particles.

        >>> job = cs.create_job("P3", "W1", "homo_abinit"
        ...     connections={"particles": ("J20", "particles_selected")}
        ...     params={"abinit_K": 3}
        ... )
        >>> job.queue()
        >>> job.status
        "queued"

    .. _CryoSPARC.find_job:
        tools.html#cryosparc.tools.CryoSPARC.find_job

    .. _Project.find_job:
        project.html#cryosparc.project.Project.find_job
    """

    def __init__(self, cs: "CryoSPARC", project_uid: str, uid: str) -> None:
        self.cs = cs
        self.project_uid = project_uid
        self.uid = uid

    @property
    def status(self) -> JobStatus:
        """
        JobStatus: scheduling status.
        """
        return self.doc["status"]

    def refresh(self):
        """
        Reload this job from the CryoSPARC database.

        Returns:
            Job: self
        """
        self._doc = self.cs.cli.get_job(self.project_uid, self.uid)  # type: ignore
        return self

    def dir(self) -> PurePosixPath:
        """
        Get the path to the job directory.

        Returns:
            Path: job directory Pure Path instance
        """
        return PurePosixPath(self.cs.cli.get_job_dir_abs(self.project_uid, self.uid))  # type: ignore

    def queue(self, lane: str, hostname: Optional[str] = None, gpus: List[int] = []):
        """
        Queue a job to a target lane. Available lanes may be queried from
        `CryoSPARC.get_lanes`_.

        Optionally specify a hostname in that lane and/or specific GPUs to use
        for computation. Available hostnames for a given lane may be queried
        with `CryoSPARC.get_targets`_.

        Args:
            lane (str): Configuried compute lane to queue to.
            hostname (str, optional): Specific hostname in compute lane, if more
                than one is available. Defaults to None.
            gpus (list[int], optional): GPUs to queue to. If specified, must
                have as many GPUs as required in job parameters. Leave
                unspecified to use first available GPU(s). Defaults to [].

        Examples:

            Queue a job to lane named "worker":

            >>> cs = CryoSPARC()
            >>> job = cs.find_job("P3", "J42")
            >>> job.status
            "building"
            >>> job.queue("worker")
            >>> job.status
            "queued"

        .. _CryoSPARC.get_lanes:
            tools.html#cryosparc.tools.CryoSPARC.get_lanes
        .. _CryoSPARC.get_targets:
            tools.html#cryosparc.tools.CryoSPARC.get_targets
        """
        self.cs.cli.enqueue_job(  # type: ignore
            project_uid=self.project_uid,
            job_uid=self.uid,
            lane=lane,
            user_id=self.cs.user_id,
            hostname=hostname,
            gpus=gpus if gpus else False,
        )
        self.refresh()

    def kill(self):
        """
        Kill this job.
        """
        self.cs.cli.kill_job(  # type: ignore
            project_uid=self.project_uid, job_uid=self.uid, killed_by_user_id=self.cs.user_id
        )
        self.refresh()

    def wait_for_status(self, status: Union[JobStatus, Iterable[JobStatus]], timeout: Optional[int] = None) -> str:
        """
        Wait for a job's status to reach the specified value. Must be one of
        the following:

        - 'building'
        - 'queued'
        - 'launched'
        - 'started'
        - 'running'
        - 'waiting'
        - 'completed'
        - 'killed'
        - 'failed'

        Args:
            status (str | set[str]): Specific status or set of statuses to wait
                for. If a set of statuses is specified, waits util job reaches
                any of the specified statuses.
            timeout (int, optional): If specified, wait at most this many
                seconds. Once timeout is reached, returns current status.
                Defaults to None.

        Returns:
            str: current job status
        """
        statuses = {status} if isinstance(status, str) else set(status)
        tic = time()
        self.refresh()
        while self.status not in statuses:
            if timeout is not None and time() - tic > timeout:
                break
            sleep(5)
            self.refresh()
        return self.status

    def wait_for_done(self, error_on_incomplete: bool = False, timeout: Optional[int] = None) -> str:
        """
        Wait until a job reaches status "completed", "killed" or "failed".

        Args:
            error_on_incomplete (bool, optional): If True, raises an assertion
                error when job finishes with status other than "completed" or
                timeout is reached. Defaults to False.
            timeout (int, optional): If specified, wait at most this many
                seconds. Once timeout is reached, returns current status or
                fails if ``error_on_incomplete`` is ``True``. Defaults to None.
        """
        status = self.wait_for_status({"completed", "killed", "failed"}, timeout=timeout)
        assert (
            not error_on_incomplete or status == "completed"
        ), f"Job {self.project_uid}-{self.uid} did not complete (status {status})"
        return status

    def clear(self):
        """
        Clear this job and reset to building status.
        """
        self.cs.cli.clear_job(self.project_uid, self.uid)  # type: ignore
        self.refresh()

    def set_param(self, name: str, value: Any, refresh: bool = True) -> bool:
        """
        Set the given param name on the current job to the given value. Only
        works if the job is in "building" status.

        Args:
            name (str): Param name, as defined in the job document's ``params_base``.
            value (any): Target parameter value.
            refresh (bool, optional): Auto-refresh job document after
                connecting. Defaults to True.

        Returns:
            bool: False if the job encountered a build error.

        Examples:

            Set the number of GPUs used by a supported job

            >>> cs = CryoSPARC()
            >>> job = cs.find_job("P3", "J42")
            >>> job.set_param("compute_num_gpus", 4)
            True
        """
        result: bool = self.cs.cli.job_set_param(  # type: ignore
            project_uid=self.project_uid, job_uid=self.uid, param_name=name, param_new_value=value
        )
        if refresh:
            self.refresh()
        return result

    def connect(self, target_input: str, source_job_uid: str, source_output: str, refresh: bool = True) -> bool:
        """
        Connect the given input for this job to an output with given job UID and
        name.

        Args:
            target_input (str): Input name to connect into. Will be created if
                not specified.
            source_job_uid (str): Job UID to connect from, e.g., "J42"
            source_output (str): Job output name to connect from , e.g.,
                "particles"
            refresh (bool, optional): Auto-refresh job document after
                connecting. Defaults to True.

        Returns:
            bool: False if the job encountered a build error.

        Examples:

            Connect J3 to CTF-corrected micrographs from J2's ``micrographs``
            output.

            >>> cs = CryoSPARC()
            >>> project = cs.find_project("P3")
            >>> job = project.find_job("J3")
            >>> job.connect("input_micrographs", "J2", "micrographs")

        """
        assert source_job_uid != self.uid, f"Cannot connect job {self.uid} to itself"
        result: bool = self.cs.cli.job_connect_group(  # type: ignore
            project_uid=self.project_uid,
            source_group=f"{source_job_uid}.{source_output}",
            dest_group=f"{self.uid}.{target_input}",
        )
        if refresh:
            self.refresh()
        return result

    def disconnect(self, target_input: str, connection_idx: Optional[int] = None, refresh: bool = True):
        """
        Clear the given job input group.

        Args:
            target_input (str): Name of input to disconnect
            connection_idx (int, optional): Connection index to clear.
                Set to 0 to clear the first connection, 1 for the second, etc.
                If unspecified, clears all connections. Defaults to None.
            refresh (bool, optional): Auto-refresh job document after
                connecting. Defaults to True.
        """
        if connection_idx is None:
            # Clear all input connections
            input_group = first(group for group in self.doc["input_slot_groups"] if group["name"] == target_input)
            if not input_group:
                raise ValueError(f"Unknown input group {target_input} for job {self.project_uid}-{self.uid}")
            for _ in input_group["connections"]:
                self.cs.cli.job_connected_group_clear(  # type: ignore
                    project_uid=self.project_uid,
                    dest_group=f"{self.uid}.{target_input}",
                    connect_idx=0,
                )
        else:
            self.cs.cli.job_connected_group_clear(  # type: ignore
                project_uid=self.project_uid,
                dest_group=f"{self.uid}.{target_input}",
                connect_idx=connection_idx,
            )

        if refresh:
            self.refresh()

    def load_input(self, name: str, slots: Iterable[str] = []):
        """
        Load the dataset connected to the job's input with the given name.

        Args:
            name (str): Input to load
            fields (list[str], optional): List of specific slots to load, such
                as ``movie_blob`` or ``locations``, or all slots if not
                specified. Defaults to [].

        Raises:
            TypeError: If the job doesn't have the given input or the dataset
                cannot be loaded.

        Returns:
            Dataset: Loaded dataset
        """
        job = self.doc
        group = first(s for s in job["input_slot_groups"] if s["name"] == name)
        if not group:
            raise TypeError(f"Job {self.project_uid}-{self.uid} does not have an input {name}")

        data = {"project_uid": self.project_uid, "job_uid": self.uid, "input_name": name, "slots": list(slots)}
        with make_json_request(self.cs.vis, "/load_job_input", data=data) as response:
            mime = response.headers.get("Content-Type")
            if mime != "application/x-cryosparc-dataset":
                raise TypeError(f"Unable to load dataset for job {self.project_uid}-{self.uid} input {name}")
            return Dataset.load(response)

    def load_output(self, name: str, slots: Iterable[str] = [], version: Union[int, Literal["F"]] = "F"):
        """
        Load the dataset for the job's output with the given name.

        Args:
            name (str): Output to load
            slots (list[str], optional): List of specific slots to load,
                such as ``movie_blob`` or ``locations``, or all slots if
                not specified (including passthrough). Defaults to [].
            version (int | Literal["F"], optional): Specific output version to
                load. Use this to load the output at different stages of
                processing. Leave unspecified to load final verion. Defaults to
                "F"

        Raises:
            TypeError: If job does not have any results for the given output

        Returns:
            Dataset: Loaded dataset
        """
        job = self.doc
        slots = set(slots)
        version = -1 if version == "F" else version
        results = [
            result
            for result in job["output_results"]
            if result["group_name"] == name and (not slots or result["name"] in slots)
        ]
        if not results:
            raise TypeError(f"Job {self.project_uid}-{self.uid} does not have any results for output {name}")

        metafiles = set(r["metafiles"][0 if r["passthrough"] else version] for r in results)
        datasets = [self.cs.download_dataset(self.project_uid, f) for f in metafiles]
        return Dataset.innerjoin(*datasets)

    def log(self, text: str, level: Literal["text", "warning", "error"] = "text"):
        """
        Append to a job's event log.

        Args:
            text (str): Text to log
            level (str, optional): Log level ("text", "warning" or "error").
                Defaults to "text".

        Returns:
            str: Created log event ID
        """
        return self.cs.cli.job_send_streamlog(  # type: ignore
            project_uid=self.project_uid, job_uid=self.uid, message=text, error=level != "text"
        )

    def log_checkpoint(self, meta: dict = {}):
        """
        Append a checkpoint to the job's event log.

        Args:
            meta (dict, optional): Additional meta information. Defaults to {}.

        Returns:
            str: Created checkpoint event ID
        """
        return self.cs.cli.job_checkpoint_streamlog(  # type: ignore
            project_uid=self.project_uid, job_uid=self.uid, meta=meta
        )

    def log_plot(
        self,
        figure: Union[str, PurePath, IO[bytes], Any],
        text: str,
        formats: Iterable[ImageFormat] = ["png", "pdf"],
        raw_data: Union[str, bytes, Literal[None]] = None,
        raw_data_file: Union[str, PurePath, IO[bytes], Literal[None]] = None,
        raw_data_format: Optional[TextFormat] = None,
        flags: List[str] = ["plots"],
        savefig_kw: dict = dict(bbox_inches="tight", pad_inches=0),
    ):
        """
        Add a log line with the given figure.

        ``figure`` must be one of the following

        - Path to an existing image file in PNG, JPEG, GIF, SVG or PDF format
        - A file handle-like object with the binary data of an image
        - A matplotlib plot

        If a matplotlib figure is specified, Uploads the plots in ``png`` and
        ``pdf`` formats. Override the ``formats`` argument with
        ``formats=['<format1>', '<format2>', ...]`` to save in different image
        formats.

        If a text-version of the given plot is available (e.g., in ``csv``
        format), specify ``raw_data`` with the full contents or
        ``raw_data_file`` with a path or binary file handle pointing to the
        contents. Assumes file format from extension or ``raw_data_format``.
        Defaults to ``"txt"`` if cannot be determined.

        Args:
            figure (str | Path | IO | Figure): Image file path, file handle or
                matplotlib figure instance
            text (str): Associated description for given figure
            formats (list[ImageFormat], optional): Image formats to save plot
                into. If a ``figure`` is a file handle, specify
                ``formats=['<format>']``, where ``<format>`` is a valid image
                extension such as ``png`` or ``pdf``. Assumes ``png`` if not
                specified. Defaults to ["png", "pdf"].
            raw_data (str | bytes, optional): Raw text data for associated plot,
                generally in CSV, XML or JSON format. Cannot be specified with
                ``raw_data_file``. Defaults to None.
            raw_data_file (str | Path | IO, optional): Path to raw text data.
                Cannot be specified with ``raw_data``. Defaults to None.
            raw_data_format (TextFormat, optional): Format for raw text data.
                Defaults to None.
            flags (list[str], optional): Flags to use for UI rendering.
                Generally should not be specified. Defaults to ["plots"].
            savefig_kw (dict, optional): If a matplotlib figure is specified
                optionally specify keyword arguments for the ``savefig`` method.
                Defaults to dict(bbox_inches="tight", pad_inches=0).

        Returns:
            str: Created log event ID
        """
        imgfiles = self.upload_plot(
            figure,
            name=text,
            formats=formats,
            raw_data=raw_data,
            raw_data_file=raw_data_file,
            raw_data_format=raw_data_format,
            savefig_kw=savefig_kw,
        )

        return self.cs.cli.job_send_streamlog(  # type: ignore
            project_uid=self.project_uid, job_uid=self.uid, message=text, flags=flags, imgfiles=imgfiles
        )

    def list_files(self, prefix: Union[str, PurePosixPath] = "", recursive: bool = False) -> List[str]:
        """
        Get a list of files inside the job directory.

        Args:
            prefix (str | Path, optional): Subdirectory inside job to list.
                Defaults to "".
            recursive (bool, optional): If True, lists files recursively.
                Defaults to False.

        Returns:
            list[str]: List of file paths relative to the job directory.
        """
        root = PurePosixPath(self.uid)
        files = self.cs.list_files(self.project_uid, prefix=root / prefix, recursive=recursive)
        return [str(PurePosixPath(f).relative_to(root)) for f in files]  # Strip leading "J#/"

    def download(self, path_rel: Union[str, PurePosixPath]):
        """
        Initiate a download request for a file inside the job's diretory

        Args:
            path_rel (str | Path): Relative path to file in job directory.

        Yields:
            HTTPResponse: Use a context manager to read the file from the
                request body.

        Examples:

            Download a job's metadata

            >>> cs = CryoSPARC()
            >>> job = cs.find_job("P3", "J42")
            >>> with job.download("job.json") as res:
            >>>     job_data = json.loads(res.read())

        """
        path_rel = PurePosixPath(self.uid) / path_rel
        return self.cs.download(self.project_uid, path_rel)

    def download_file(self, path_rel: Union[str, PurePosixPath], target: Union[str, PurePath, IO[bytes]]):
        """
        Download file from job directory to the given target path or writeable
        file handle.

        Args:
            path_rel (str | Path): Relative path to file in job directory.
            target (str | Path | IO): Local file path, directory path or writeable
                file handle to write response data.

        Returns:
            Path | IO: resulting target path or file handle.
        """
        path_rel = PurePosixPath(self.uid) / path_rel
        return self.cs.download_file(self.project_uid, path_rel, target)

    def download_dataset(self, path_rel: Union[str, PurePosixPath]):
        """
        Download a .cs dataset file from the given relative path in the job
        directory.

        Args:
            path_rel (str | Path): Relative path to .cs file in job directory.

        Returns:
            Dataset: Loaded dataset instance
        """
        path_rel = PurePosixPath(self.uid) / path_rel
        return self.cs.download_dataset(self.project_uid, path_rel)

    def download_mrc(self, path_rel: Union[str, PurePosixPath]):
        """
        Download a .mrc file from the given relative path in the job directory.

        Args:
            path (str | Path): Relative path to .mrc file in job directory.

        Returns:
            tuple[Header, NDArray]: MRC file header and data as a numpy array
        """
        path_rel = PurePosixPath(self.uid) / path_rel
        return self.cs.download_mrc(self.project_uid, path_rel)

    def list_assets(self) -> List[AssetDetails]:
        """
        Get a list of files available in the database for this job. Returns a
        list with details about the assets. Each entry is a dict with a ``_id``
        key which may be used to download the file with the ``download_asset``
        method.

        Returns:
            list[AssetDetails]: Asset details
        """
        return self.cs.vis.list_job_files(project_uid=self.project_uid, job_uid=self.uid)  # type: ignore

    def download_asset(self, fileid: str, target: Union[str, PurePath, IO[bytes]]):
        """
        Download a job asset from the database with the given ID. Note that the
        file does not necessary have to belong to the current job.

        Args:
            fileid (str): GridFS file object ID
            target (str | Path | IO): Local file path, directory path or
                writeable file handle to write response data.

        Returns:
            Path | IO: resulting target path or file handle.

        """
        return self.cs.download_asset(fileid, target)

    def upload(self, target_path_rel: Union[str, PurePosixPath], source: Union[str, bytes, PurePath, IO]):
        """
        Upload the given file to the job directory at the given path.

        Args:
            target_path_rel (str | Path): Relative target path in job directory
            source (str | bytes | Path | IO): Local path or file handle to
                upload. May also specified as raw bytes.
        """
        target_path_rel = PurePosixPath(self.uid) / target_path_rel
        return self.cs.upload(self.project_uid, target_path_rel, source)

    def upload_asset(
        self,
        file: Union[str, PurePath, IO[bytes]],
        filename: Optional[str] = None,
        format: Optional[AssetFormat] = None,
    ) -> EventLogAsset:
        """
        Upload an image or text file to the current job. Specify either an image
        (PNG, JPG, GIF, PDF, SVG), text file (TXT, CSV, JSON, XML) or a binary
        IO object with data in one of those formats.

        If a binary IO object is specified, either a filename or mimetype must
        be specified.

        Unlike the ``upload`` method which saves files to the job directory,
        this method saves images to the database and exposes them for use in the
        job log.

        If specifying arbitrary binary I/O, specify either a filename or a file
        format.

        Args:
            file (str | Path | IO): Source asset file path or handle
            filename (str, optional): Filename of asset. If ``file`` is a handle
                specify one of ``filename`` or ``format``. Defaults to None.
            format (AssetFormat, optional): Format of filename. If ``file`` is
                a handle, specify one of ``filename`` or ``format``. Defaults to
                None.

        Raises:
            ValueError: If incorrect arguments specified

        Returns:
            EventLogAsset: Dictionary including details about uploaded asset.
        """
        if format:
            assert format in ASSET_CONTENT_TYPES, f"Invalid asset format {format}"
        elif filename:
            ext = filename.split(".")[-1]
            assert ext in ASSET_CONTENT_TYPES, f"Invalid asset format {ext}"
            format = ext
        elif isinstance(file, (str, PurePath)):
            file = PurePath(file)
            filename = file.name
            ext = filename.split(".")[-1]
            assert ext in ASSET_CONTENT_TYPES, f"Invalid asset format {ext}"
            format = ext
        else:
            raise ValueError("Must specify filename or format when saving binary asset handle")

        with bopen(file) as f:
            url = f"/projects/{self.project_uid}/jobs/{self.uid}/files"
            query = {"format": format}
            if filename:
                query["filename"] = filename

            with make_request(self.cs.vis, url=url, query=query, data=f) as res:
                assert res.status >= 200 and res.status < 300, (
                    f"Could not upload project {self.project_uid} asset {file}.\n"
                    f"Response from CryoSPARC: {res.read().decode()}"
                )
                return json.loads(res.read())

    def upload_plot(
        self,
        figure: Union[str, PurePath, IO[bytes], Any],
        name: Optional[str] = None,
        formats: Iterable[ImageFormat] = ["png", "pdf"],
        raw_data: Union[str, bytes, Literal[None]] = None,
        raw_data_file: Union[str, PurePath, IO[bytes], Literal[None]] = None,
        raw_data_format: Optional[TextFormat] = None,
        savefig_kw: dict = dict(bbox_inches="tight", pad_inches=0),
    ) -> List[EventLogAsset]:
        """
        Upload the given figure. Returns a list of the created asset objects.
        Avoid using directly; use ``log_plot`` instead. See ``log_plot``
        additional details.

        Args:
            figure (str | Path | IO | Figure): Image file path, file handle or
                matplotlib figure instance
            name (str): Associated name for given figure
            formats (list[ImageFormat], optional): Image formats to save plot
                into. If a ``figure`` is a file handle, specify
                ``formats=['<format>']``, where ``<format>`` is a valid image
                extension such as ``png`` or ``pdf``. Assumes ``png`` if not
                specified. Defaults to ["png", "pdf"].
            raw_data (str | bytes, optional): Raw text data for associated plot,
                generally in CSV, XML or JSON format. Cannot be specified with
                ``raw_data_file``. Defaults to None.
            raw_data_file (str | Path | IO, optional): Path to raw text data.
                Cannot be specified with ``raw_data``. Defaults to None.
            raw_data_format (TextFormat, optional): Format for raw text data.
                Defaults to None.
            savefig_kw (dict, optional): If a matplotlib figure is specified
                optionally specify keyword arguments for the ``savefig`` method.
                Defaults to dict(bbox_inches="tight", pad_inches=0).

        Raises:
            ValueError: If incorrect argument specified

        Returns:
            list[EventLogAsset]: Details about created uploaded job assets
        """
        figdata = []
        basename = name or "figure"
        if hasattr(figure, "savefig"):  # matplotlib plot
            for fmt in formats:
                assert fmt in IMAGE_CONTENT_TYPES, f"Invalid figure format {fmt}"
                filename = f"{basename}.{fmt}"
                data = BytesIO()
                figure.savefig(data, format=fmt, **savefig_kw)  # type: ignore
                data.seek(0)
                figdata.append((data, filename, fmt))
        elif isinstance(figure, (str, PurePath)):  # file path; assume format from filename
            figure = PurePath(figure)
            basename = figure.stem
            fmt = str(figure).split(".")[-1]
            assert fmt in IMAGE_CONTENT_TYPES, f"Invalid figure format {fmt}"
            filename = f"{name or figure.stem}.{fmt}"
            figdata.append((figure, filename, fmt))
        else:  # Binary IO
            fmt = first(iter(formats))
            assert fmt in IMAGE_CONTENT_TYPES, f"Invalid or unspecified figure format {fmt}"
            filename = f"{basename}.{fmt}"
            figdata.append((figure, filename, fmt))

        raw_data_filename = f"{name}.{raw_data_format or 'txt'}" if name else None
        if raw_data and raw_data_file:
            raise ValueError("Ambiguous invocation; specify only one of raw_data or raw_data_file")
        elif isinstance(raw_data, str):
            raw_data_file = BytesIO(raw_data.encode())
        elif isinstance(raw_data, bytes):
            raw_data_file = BytesIO(raw_data)
        elif isinstance(raw_data_file, (str, PurePath)):
            raw_data_path = PurePath(raw_data_file)
            raw_data_filename = raw_data_path.name
            ext = raw_data_format or raw_data_filename.split(".")[-1]
            assert ext in TEXT_CONTENT_TYPES, f"Invalid raw data filename {raw_data_file}"
            raw_data_format = ext

        assets = []
        for data, filename, fmt in figdata:
            asset = self.upload_asset(data, filename=filename, format=fmt)
            assets.append(asset)

        if raw_data_file:
            raw_data_format = raw_data_format or "txt"
            asset = self.upload_asset(raw_data_file, filename=raw_data_filename, format=raw_data_format)
            assets.append(asset)

        return assets

    def upload_dataset(self, target_path_rel: Union[str, PurePosixPath], dset: Dataset, format: int = DEFAULT_FORMAT):
        """
        Upload a dataset as a CS file into the job directory.

        Args:
            target_path_rel (str | Path): relative path to save dataset in job
                directory. Should have a ``.cs`` extension.
            dset (Dataset): dataset to save.
            format (int): format to save in from ``cryosparc.dataset.*_FORMAT``,
                defaults to NUMPY_FORMAT)

        """
        target_path_rel = PurePosixPath(self.uid) / target_path_rel
        return self.cs.upload_dataset(self.project_uid, target_path_rel, dset, format=format)

    def upload_mrc(self, target_path_rel: Union[str, PurePosixPath], data: "NDArray", psize: float):
        """
        Upload a numpy 2D or 3D array to the job directory as an MRC file.

        Args:
            target_path_rel (str | Path): relative path to save array in job
                directory. Should have ``.mrc`` extension.
            data (NDArray): Numpy array with MRC file data.
            psize (float): Pixel size to include in MRC header.
        """
        target_path_rel = PurePosixPath(self.uid) / target_path_rel
        return self.cs.upload_mrc(self.project_uid, target_path_rel, data, psize)

    def mkdir(
        self,
        target_path_rel: Union[str, PurePosixPath],
        parents: bool = False,
        exist_ok: bool = False,
    ):
        """
        Create a directory in the given job.

        Args:
            target_path_rel (str | Path): Relative path to create inside project
                directory.
            parents (bool, optional): If True, any missing parents are created
                as needed. Defaults to False.
            exist_ok (bool, optional): If True, does not raise an error for
                existing directories. Still raises if the target path is not a
                directory. Defaults to False.
        """
        self.cs.mkdir(
            project_uid=self.project_uid,
            target_path_rel=PurePosixPath(self.uid) / target_path_rel,
            parents=parents,
            exist_ok=exist_ok,
        )

    def cp(self, source_path_rel: Union[str, PurePosixPath], target_path_rel: Union[str, PurePosixPath]):
        """
        Copy a file or folder within a project to another location within that
        same project. Note that argument order is reversed from
        equivalent ``cp`` command.

        Args:
            source_path_rel (str | Path): Relative path in project of source
                file or folder to copy.
            target_path_rel (str | Path): Relative path in project to copy to.
        """
        self.cs.cp(
            project_uid=self.project_uid,
            source_path_rel=PurePosixPath(self.uid) / source_path_rel,
            target_path_rel=PurePosixPath(self.uid) / target_path_rel,
        )

    def symlink(self, source_path_rel: Union[str, PurePosixPath], target_path_rel: Union[str, PurePosixPath]):
        """
        Create a symbolic link in the given project. May only create links for
        files within the project. Note that argument order is reversed from
        ``ln -s``.

        Args:
            project_uid (str): Target project UID, e.g., "P3".
            source_path_rel (str | Path): Relative path in project to file from
                which to create symlink.
            target_path_rel (str | Path): Relative path in project to new
                symlink.
        """
        self.cs.symlink(
            project_uid=self.project_uid,
            source_path_rel=PurePosixPath(self.uid) / source_path_rel,
            target_path_rel=PurePosixPath(self.uid) / target_path_rel,
        )

    def subprocess(
        self,
        args: Union[str, list],
        mute: bool = False,
        checkpoint: bool = False,
        checkpoint_line_pattern: Union[str, Pattern[str], Literal[None]] = None,
        **kwargs,
    ):
        """
        Launch a subprocess and write its text-based output and error to the job
        log.

        Args:
            args (str | list): Process arguments to run
            mute (bool, optional): If True, does not also forward process output
                to standard output. Defaults to False.
            checkpoint (bool, optional): If True, creates a checkpoint in the
                job event log just before process output begins. Defaults to
                False.
            checkpoint_line_pattern (str | Pattern[str], optional): Regular
                expression to match checkpoint lines for processes with a lot of
                output. If a process outputs a line that matches this pattern, a
                checkpoint is created in the event log before this line is
                forwarded. Defaults to None.
            **kwargs: Additional keyword arguments for ``subprocess.Popen``.

        Raises:
            TypeError: For invalid arguments
            RuntimeError: If process exists with non-zero status code
        """
        import subprocess
        import re

        pttrn = None
        if checkpoint_line_pattern and isinstance(checkpoint_line_pattern, str):
            pttrn = re.compile(checkpoint_line_pattern)
        elif isinstance(checkpoint_line_pattern, re.Pattern):
            pttrn = checkpoint_line_pattern
        elif checkpoint_line_pattern:
            raise TypeError(f"Invalid checkpoint_line_pattern argument type: {type(checkpoint_line_pattern)}")

        args = args if isinstance(args, str) else list(map(str, args))
        with subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True, **kwargs) as proc:
            assert proc.stdout, f"Subprocess {args} has not standard output"
            if checkpoint:
                self.log_checkpoint()

            self.log("======= Forwarding subprocess output for the following command =======")
            self.log(str(args))
            self.log("======================================================================")

            for line in proc.stdout:
                line = line.rstrip()
                if pttrn and pttrn.match(line):
                    self.log_checkpoint()
                self.log(line)
                if not mute:
                    print(line)

            while proc.poll() is None:
                sleep(1)

            if proc.returncode != 0:
                msg = f"Subprocess {args} exited with status {proc.returncode}"
                self.log(msg, level="error")
                raise RuntimeError(msg)

            self.log("======================= Subprocess complete. =========================")


class ExternalJob(Job):
    """
    Mutable custom output job with customizeble input slots and output results.
    Use External jobs to save data save cryo-EM data generated by a software
    package outside of CryoSPARC.

    Created external jobs may be connected to any other CryoSPARC job result as
    an input. Its outputs must be created manually and may be configured to
    passthrough inherited input fields, just as with regular CryoSPARC jobs.

    Create a new External Job with `Project.create_external_job`_.

    Examples:

        Import multiple exposure groups into a single job

        >>> from cryosparc.tools import CryoSPARC
        >>> cs = CryoSPARC()
        >>> project = cs.find_project("P3")
        >>> job = project.create_external_job("W3", title="Import Image Sets")
        >>> for i in range(3):
        ...     dset = job.add_output(
        ...         type="exposure",
        ...         name=f"images_{i}",
        ...         slots=["movie_blob", "mscope_params", "gain_ref_blob"],
        ...         alloc=10  # allocate a dataset for this output with 10 rows
        ...     )
        ...     dset['movie_blob/path'] = ...  # populate dataset
        ...     job.save_output(output_name, dset)

    .. _Project.create_external_job:
        project.html#cryosparc.project.Project.create_external_job

    """

    def add_input(
        self,
        type: Datatype,
        name: Optional[str] = None,
        min: int = 0,
        max: Union[int, Literal["inf"]] = "inf",
        slots: Iterable[Union[str, Datafield]] = [],
        title: Optional[str] = None,
    ):
        """
        Add an input slot to the current job. May be connected to zero or more
        outputs from other jobs (depending on the min and max values).

        Args:
            type (Datatype): cryo-EM data type for this output, e.g., "particle"
            name (str, optional): Output name key, e.g., "picked_particles".
                Defaults to None.
            min (int, optional): Minimum number of required input connections.
                Defaults to 0.
            max (int | Literal["inf"], optional): Maximum number of input
                connections. Specify ``"inf"`` for unlimited connections.
                Defaults to "inf".
            slots (list[str | Datafield], optional): List of slots that should
                be connected to this input, such as ``"location"`` or ``"blob"``
                Defaults to [].
            title (str, optional): Human-readable title for this input. Defaults
                to None.

        Returns:
            str: name of created input

        Examples:

            Create an external job that accepts micrographs as input:

            >>> cs = CryoSPARC()
            >>> project = cs.find_project("P3")
            >>> job = project.create_external_job("W1", title="Custom Picker")
            >>> job.uid
            "J3"
            >>> job.add_input(
            ...     type="exposure",
            ...     name="input_micrographs",
            ...     min=1,
            ...     slots=["micrograph_blob", "ctf"],
            ...     title="Input micrographs for picking
            ... )
            "input_micrographs"
        """
        self.cs.vis.add_external_job_input(  # type: ignore
            project_uid=self.project_uid,
            job_uid=self.uid,
            type=type,
            name=name,
            min=min,
            max=max,
            slots=slots,
            title=title,
        )
        self.refresh()
        return self.doc["input_slot_groups"][-1]["name"]

    @overload
    def add_output(
        self,
        type: Datatype,
        name: Optional[str] = ...,
        slots: List[Union[str, Datafield]] = ...,
        passthrough: Optional[str] = ...,
        title: Optional[str] = None,
    ) -> str:
        ...

    @overload
    def add_output(
        self,
        type: Datatype,
        name: Optional[str] = ...,
        slots: List[Union[str, Datafield]] = ...,
        passthrough: Optional[str] = ...,
        title: Optional[str] = None,
        alloc: Union[int, Dataset] = ...,
    ) -> Dataset:
        ...

    def add_output(
        self,
        type: Datatype,
        name: Optional[str] = None,
        slots: List[Union[str, Datafield]] = [],
        passthrough: Optional[str] = None,
        title: Optional[str] = None,
        alloc: Union[int, Dataset, Literal[None]] = None,
    ) -> Union[str, Dataset]:
        """
        Add an output slot to the current job. Optionally returns the
        corresponding empty dataset if ``alloc`` is specified.

        Args:
            type (Datatype): cryo-EM datatype for this output, e.g., "particle"
            name (str, optional): Output name key, e.g., "selected_particles".
                Same as ``type`` if not specified. Defaults to None.
            slots (list[str, Datafield], optional): List of slot expected to be
                created for this output, such as ``location`` or ``blob``. Do
                not specify any slots that were passed through from an input
                unless those slots are modified in the output. Defaults to [].
            passthrough (str, optional): Indicates that this output inherits
                slots from an existing input with the specified name. The input
                must first be added with ``add_input()``. Defaults to False.
            title (str, optional): Human-readable title for this input. Defaults
                to None.
            alloc (int | Dataset, optional): If specified, pre-allocate and
                return a dataset with the requested slots. Specify an integer
                to allocate a specific number of rows. Specify a Dataset from
                which to inherit unique row IDs (useful when adding passthrough
                outputs). Defaults to None.

        Returns:
            str | Dataset: Name of the created output. If ``alloc`` is
                specified as an integer, instead returns blank dataset with the
                given size and random UIDs. If ``alloc`` is specified as a
                Dataset, returns blank dataset with the same UIDs.

        Examples:

            Create and allocate an output for new particle picks

            >>> cs = CryoSPARC()
            >>> project = cs.find_project("P3")
            >>> job = project.find_external_job("J3")
            >>> particles_dset = job.add_output(
            ...     type="particle",
            ...     name="picked_particles",
            ...     slots=["location", "pick_stats"],
            ...     alloc=10000
            ... )

            Create an inheritied output for input micrographs

            >>> job.add_output(
            ...     type="exposures",
            ...     name="picked_micrographs",
            ...     passthrough="input_micrographs",
            ...     title="Passthrough picked micrographs"
            ... )
            "picked_micrographs"

            Create an output with multiple slots of the same type

            >>> job.add_output(
            ...     type="particle",
            ...     name="particle_alignments",
            ...     slots=[
            ...         {"dtype": "alignments3D", "prefix": "alignments_class_0", "required": True},
            ...         {"dtype": "alignments3D", "prefix": "alignments_class_1", "required": True},
            ...         {"dtype": "alignments3D", "prefix": "alignments_class_2", "required": True},
            ...     ]
            ... )
            "particle_alignments"
        """
        self.cs.vis.add_external_job_output(  # type: ignore
            project_uid=self.project_uid,
            job_uid=self.uid,
            type=type,
            name=name,
            slots=slots,
            passthrough=passthrough,
            title=title,
        )
        self.refresh()
        result_name = self.doc["output_result_groups"][-1]["name"]
        return result_name if alloc is None else self.alloc_output(result_name, alloc)

    def connect(
        self,
        target_input: str,
        source_job_uid: str,
        source_output: str,
        slots: List[Union[str, Datafield]] = [],
        title: str = "",
        desc: str = "",
        refresh: bool = True,
    ):
        """
        Connect the given input for this job to an output with given job UID and
        name. If this input does not exist, it will be added with the given
        slots. At least one slot must be specified if the input does not exist.

        Args:
            target_input (str): Input name to connect into. Will be created if
                does not already exist.
            source_job_uid (str): Job UID to connect from, e.g., "J42"
            source_output (str): Job output name to connect from , e.g.,
                "particles"
            slots (list[str | Datafield], optional): List of slots to add to
                created input. All if not specified. Defaults to [].
            title (str, optional): Human readable title for created input.
                Defaults to "".
            desc (str, optional): Human readable description for created input.
                Defaults to "".
            refresh (bool, optional): Auto-refresh job document after
                connecting. Defaults to True.

        Examples:

            Connect J3 to CTF-corrected micrographs from J2's ``micrographs``
            output.

            >>> cs = CryoSPARC()
            >>> project = cs.find_project("P3")
            >>> job = project.find_external_job("J3")
            >>> job.connect("input_micrographs", "J2", "micrographs")

        """
        assert source_job_uid != self.uid, f"Cannot connect job {self.uid} to itself"
        self.cs.vis.connect_external_job(  # type: ignore
            project_uid=self.project_uid,
            source_job_uid=source_job_uid,
            source_output=source_output,
            target_job_uid=self.uid,
            target_input=target_input,
            slots=slots,
            title=title,
            desc=desc,
        )
        if refresh:
            self.refresh()

    def alloc_output(self, name: str, alloc: Union[int, "ArrayLike", Dataset] = 0) -> Dataset:
        """
        Allocate an empty dataset for the given output with the given name.
        Initialize with the given number of empty rows. The result may be
        used with ``save_output`` with the same output name.

        Args:
            name (str): Name of job output to allocate
            size (int | ArrayLike | Dataset, optional): Specify as one of the
                following: (A) integer to allocate a specific number of rows,
                (B) a numpy array of numbers to use for UIDs in the allocated
                dataset or (C) a dataset from which to inherit unique row IDs
                (useful  for allocating passthrough outputs). Defaults to 0.

        Returns:
            Dataset: Empty dataset with the given number of rows

        Examples:

            Allocate a dataset of size 10,000 for an output for new particle
            picks

            >>> cs = CryoSPARC()
            >>> project = cs.find_project("P3")
            >>> job = project.find_external_job("J3")
            >>> job.alloc_output("picked_particles", 10000)
            Dataset([  # 10000 items, 11 fields
                ("uid": [...]),
                ("location/micrograph_path", ["", ...]),
                ...
            ])

            Allocate a dataset from an existing input passthrough dataset

            >>> input_micrographs = job.load_input("input_micrographs")
            >>> job.alloc_output("picked_micrographs", input_micrographs)
            Dataset([  # same "uid" field as input_micrographs
                ("uid": [...]),
            ])

        """
        expected_fields = []
        for result in self.doc["output_results"]:
            if result["group_name"] != name or result["passthrough"]:
                continue
            prefix = result["name"]
            for field, dtype in result["min_fields"]:
                expected_fields.append((f"{prefix}/{field}", dtype))

        if not expected_fields:
            raise ValueError(f"No such output {name} on {self.project_uid}-{self.uid}")

        if isinstance(alloc, int):
            return Dataset.allocate(alloc, expected_fields)
        elif isinstance(alloc, Dataset):
            return Dataset({"uid": alloc["uid"]}).add_fields(expected_fields)
        else:
            return Dataset({"uid": alloc}).add_fields(expected_fields)

    def save_output(self, name: str, dataset: Dataset, refresh: bool = True):
        """
        Save output dataset to external job.

        Args:
            name (str): Name of output on this job.
            dataset (Dataset): Value of output with only required fields.
            refresh (bool, Optional): Auto-refresh job document after saving.
                Defaults to True

        Examples:

            Save a previously-allocated output.

            >>> cs = CryoSPARC()
            >>> project = cs.find_project("P3")
            >>> job = project.find_external_job("J3")
            >>> particles = job.alloc_output("picked_particles", 10000)
            >>> job.save_output("picked_particles", particles)

        """
        url = f"/external/projects/{self.project_uid}/jobs/{self.uid}/outputs/{name}/dataset"
        with make_request(self.cs.vis, url=url, data=dataset.stream()) as res:
            result = res.read().decode()
            assert res.status >= 200 and res.status < 400, f"Save output failed with message: {result}"
        if refresh:
            self.refresh()

    def start(self, status: Literal["running", "waiting"] = "waiting"):
        """
        Set job status to "running" or "waiting"

        Args:
            status (str, optional): "running" or "waiting". Defaults to "waiting".
        """
        assert status in {"running", "waiting"}, f"Invalid start status {status}"
        assert self.doc["status"] not in {
            "running",
            "waiting",
        }, f"Job {self.project_uid}-{self.uid} is already in running status"
        self.cs.cli.run_external_job(self.project_uid, self.uid, status)  # type: ignore
        self.refresh()

    def stop(self, error=False):
        """
        Set job status to "completed" or "failed"

        Args:
            error (bool, optional): Job completed with errors. Defaults to False.
        """
        status = "failed" if error else "completed"
        self.cs.cli.set_job_status(self.project_uid, self.uid, status)  # type: ignore
        self.refresh()

    @contextmanager
    def run(self):
        """
        Start a job within a context manager and stop the job when the context
        ends.

        Yields:
            ExternalJob: self.

        Examples:

            Job will be marked as "failed" if the contents of the block throw an
            exception

            >>> with job.run():
            ...     job.save_output(...)

        """
        error = False
        self.start("running")
        self.refresh()
        try:
            yield self
        except Exception:
            error = True
            raise
        finally:
            self.stop(error)  # TODO: Write Error to job log, if possible
            self.refresh()



"""
Type specifications for CryoSPARC database entities.

Unless otherwise noted, classes defined here represent dictionary instances
whose attributes may be accessed with dictionary key syntax.

Examples:

    Accessing job document details

    >>> cs = CryoSPARC()
    >>> job = cs.find_job("P3", "J118")
    >>> job.doc["output_results"][0]["metafiles"]
    [
      "J118/J118_000_particles.cs",
      "J118/J118_001_particles.cs",
      "J118/J118_002_particles.cs",
      "J118/J118_003_particles.cs"
    ]
"""
from abc import ABC, abstractmethod
from typing import Any, Dict, Generic, List, Optional, Tuple, TypeVar, Union
from typing_extensions import Literal, TypedDict

# Database document
D = TypeVar("D", bound=TypedDict)


Datatype = Literal[
    "exposure", "particle", "template", "volume", "mask", "ml_model", "symmetry_candidate", "flex_mesh", "flex_model"
]
"""Supported data types for job inputs and outputs."""


JobStatus = Literal[
    "building",
    "queued",
    "launched",
    "started",
    "running",
    "waiting",
    "completed",
    "killed",
    "failed",
]
"""
Possible job status values.
"""

# Valid plot file types
TextFormat = Literal["txt", "csv", "html", "json", "xml"]
"""
Supported job stream log asset file text formats.
"""

ImageFormat = Literal["pdf", "gif", "jpg", "jpeg", "png", "svg"]
"""
Supported job stream log asset file image formats.
"""

AssetFormat = Union[TextFormat, ImageFormat]
"""
Supported job stream log asset file formats.
"""

TextContentType = Literal[
    "text/plain",
    "text/csv",
    "text/html",
    "application/json",
    "application/xml",
]
"""
Supported job stream log text asset MIME types.
"""

ImageContentType = Literal[
    "application/pdf",
    "image/gif",
    "image/jpeg",
    "image/png",
    "image/svg+xml",
]
"""
Supported job image asset MIME types.
"""

AssetContentType = Union[TextContentType, ImageContentType]
"""
Supported job asset MIME types.
"""

TEXT_CONTENT_TYPES: Dict[TextFormat, TextContentType] = {
    "txt": "text/plain",
    "csv": "text/csv",
    "html": "text/html",
    "json": "application/json",
    "xml": "application/xml",
}

IMAGE_CONTENT_TYPES: Dict[ImageFormat, ImageContentType] = {
    "pdf": "application/pdf",
    "gif": "image/gif",
    "jpeg": "image/jpeg",
    "jpg": "image/jpeg",
    "png": "image/png",
    "svg": "image/svg+xml",
}

ASSET_CONTENT_TYPES: Dict[AssetFormat, AssetContentType] = {**TEXT_CONTENT_TYPES, **IMAGE_CONTENT_TYPES}  # type: ignore
ASSET_EXTENSIONS = {v: k for k, v in ASSET_CONTENT_TYPES.items()}


class AssetDetails(TypedDict):
    """
    Result of job asset files query. Keys may be accessed with dictionary
    syntax.

    Examples:

        >>> print(details['filename'])
        image.png
    """

    _id: str
    """Document ID"""

    filename: str
    """File name"""

    contentType: AssetContentType
    """Asset content type, e.g., "image/png" """

    uploadDate: str  # ISO formatted
    """ISO 8601-formatted asset upload date"""

    length: int  # in bytes
    """Size of file in bytes"""

    chunkSize: int  # in bytes
    """File chunk size in bytes"""

    md5: str
    """MD5 hash of asset"""

    project_uid: str
    """Associated project UID"""

    job_uid: str  # also used for Session UID
    """Associated job or session UID"""


class EventLogAsset(TypedDict):
    """
    Dictionary item in a job event log's ``imgfiles`` property (in the
    ``events`` collection). Keys may be accessed with dictionary key syntax.

    Examples:

        >>> print(asset['filename'])
        image.png
    """

    fileid: str
    """Reference to file ``_id`` property in GridFS collection"""

    filename: str
    """File name"""

    filetype: AssetContentType
    """File content type, e.g., "image/png" """


class Datafield(TypedDict):
    """
    Definition of a prefix field within a CS file.

    Examples:

        >>> field = Datafield(dtype='alignments3D', prefix='alignments_class_0', required=False)
        >>> print(field['dtype'])
        alignments3D
    """

    dtype: str
    """Datatype-specific string from based on entry in
    ``cryosparc_compute/jobs/common.py``. e.g., "movie_blob", "ctf",
    "alignments2D"."""

    prefix: str
    """where to find field in an associated ``.cs`` file. e.g.,
    "alignments_class_1" """

    required: bool
    """whether this field must necessarily exist in a corresponding
    input/output. Assumed to be ``True`` if not specified"""


class InputSlot(TypedDict):
    """
    Dictionary entry in Job document's ``input_slot_groups.slots`` property.
    """

    type: Datatype
    """Cryo-EM native data type, e.g., "exposure", "particle" or "volume" """

    name: str
    """Input slot name, e.g., "movie_blob" or "location" """

    title: str
    """Human-readable input slot title"""

    description: str
    """Human-readable description"""

    optional: bool
    """If True, input is not required for the job"""


class ConnectionSlot(TypedDict):
    """
    Slots specified entry in a Job document's ``input_slot_groups[].connections[].slots`` list.
    """

    slot_name: Optional[str]
    """Passthrough slots have ``slot_name`` set to ``None``."""

    job_uid: str
    """Parent job UID source of this input slot connection."""

    group_name: str
    """Name of output group in parent job. e.g., "particles" """

    result_name: str
    """Name of output slot in parent job, e.g., "blob" """

    result_type: str
    """Type of result slot based on entry in ``cryosparc_compute/jobs/common.py``, e.g., "particle.blob" """

    version: Union[int, Literal["F"]]
    """Version number or specifier to use. Usually "F" """


class Connection(TypedDict):
    """
    Connection element specified in a Job document's ``input_slot_groups[].connections`` list.
    """

    job_uid: str
    """Parent job UID source of main input group connection."""

    group_name: str
    """Name of output group in parent job. e.g., "particles" """

    slots: List[ConnectionSlot]
    """List of connection specifiers for each slot"""


class InputSlotGroup(TypedDict):
    """Element specified in a Job document's ``input_slot_groups`` list."""

    type: Datatype
    """Possible Cryo-EM data type for this group, e.g., "particle"."""

    name: str
    """Input group name, e.g., "particles"."""

    title: str
    """Human-readable input group title."""

    description: str
    """Human-readable input group description."""

    count_min: int
    """Minimum required output groups that may be connected to this input slot."""

    count_max: Optional[int]
    """Maximum allowed output groups that may be connected to this input slot. Infinity if not specified."""

    repeat_allowed: bool
    """If True, the same output group may be connected twice."""

    slots: List[InputSlot]
    """List of slot definitions in the input group."""

    connections: List[Connection]
    """Connected output for this input group."""


class OutputResultGroupContains(TypedDict):
    """
    Elements of a Job document's ``output_result_groups[].contains`` list.
    """

    uid: str
    """Result unique ID, e.g., "J42-R1"."""

    type: str
    """Result type based on entry in ``cryosparc_compute/jobs/common.py``, e.g., "particle.alignments3D"."""

    name: str
    """Name of output result (a.k.a. slot), e.g., "alignments_class_1"."""

    group_name: str
    """Name of output group, e.g., "particles"."""

    passthrough: bool
    """If True, this result is passed through as-is from an associated input."""


class OutputResultGroup(TypedDict):
    """
    Elements of a Job document's ``output_result_groups`` list.
    """

    uid: str
    """Ouptut group unique ID, e.g., "J42-G1"."""

    type: Datatype
    """Possible Cryo-EM data type for this group, e.g., "particle"."""

    name: str
    """Output group name, e.g., "particles_selected" """

    title: str
    """Human-readable output group title."""

    description: str
    """Human-readable output group description."""

    contains: List[OutputResultGroupContains]
    """List of specific results (a.k.a. slots) in this output group."""

    passthrough: Union[str, Literal[False]]
    """Either ``False`` if this is a newly-created output or the name of an
    input group used to forward passthrough slots for this result group."""

    num_items: int
    """Number of rows in the dataset for this result group populated by jobs when they run."""

    summary: dict
    """Context-specific details about this result populated by jobs when they run."""


class OutputResult(TypedDict):
    """
    Detailed schema and metadata for a Job document's ``output_results`` list.
    Similar to a flattened ``output_result_groups[].contains`` but with more
    details.
    """

    uid: str
    """Result unique ID, e.g., "J42-R1"."""

    type: str
    """Result type based on entry in ``cryosparc_compute/jobs/common.py``, e.g., "particle.alignments3D"."""

    name: str
    """Name of output result (a.k.a. slot), e.g., "alignments_class_1"."""

    group_name: str
    """Name of output group, e.g., "particles"."""

    title: str
    """Human-readable output result title."""

    description: str
    """Human-readable output result description."""

    versions: List[int]
    """List of available intermediate result version numbers."""

    metafiles: List[str]
    """List of available intermediate result files (same size as ``versions``)."""

    min_fields: List[Tuple[str, str]]
    """Minimum included dataset field definitions in this result."""

    num_items: int
    """Number of rows in the dataset for this result populated by jobs when they run."""

    passthrough: bool
    """If True, this result is passed through as-is from an associated input."""


class BaseParam(TypedDict):
    """
    Base parameter specification.
    """

    value: bool
    """Base parameter value. Should not be changed."""

    title: str
    """Human-readable parameter title."""

    desc: str
    """Human-readable parameter description."""

    order: int
    """Parameter order in the builder list."""

    section: str
    """Parameter section identifier."""

    advanced: bool
    """True if this is an advanced parameter (hidden unlesss the "Advanced"
    checkbox is enabled in the Job Builder"."""

    hidden: bool
    """If True, this parameter is always hidden from the interface."""


class Param(BaseParam):
    """
    Specifies possible values for type property. Inherits from
    BaseParam_.

    .. _BaseParam:
        #cryosparc.spec.BaseParam
    """

    type: Literal["number", "string", "boolean"]
    """Possible Parameter type."""


class EnumParam(BaseParam):
    """
    Additional Param keys available for enum params. Inherits from BaseParam_.

    .. _BaseParam:
        #cryosparc.spec.BaseParam
    """

    type: Literal["enum"]
    """Possible Parameter type."""

    enum_keys: List[str]
    """Possible enum names for display for selection. Parameter must be set to
    one of these values."""

    enum_dict: Dict[str, Any]
    """Map from enum key names to their equivalent values."""


class PathParam(BaseParam):
    """
    Additional Param key available for path params. Inherits Inherits from
    BaseParam_.

    .. _BaseParam:
        #cryosparc.spec.BaseParam
    """

    type: Literal["path"]

    path_dir_allowed: bool
    """If True, directories may be specified."""

    path_file_allowed: bool
    """If True, files may be specified."""

    path_glob_allowed: bool
    """If True, a wildcard string that refers to many files may be specified.."""


class ParamSpec(TypedDict):
    """Param specification. Dictionary with single ``"value"`` key."""

    value: Any
    """Value of param."""


class ProjectLastAccessed(TypedDict, total=False):
    """
    Details on when a project was last accessed.
    """

    name: str
    """User account name that accessed this project."""

    accessed_at: str
    """Last access date in ISO 8601 format."""


class ProjectDocument(TypedDict):
    """
    Specification for a project document in the MongoDB database.
    """

    _id: str
    """MongoDB ID"""

    uid: str
    """Project unique ID, e.g., "J42"."""

    uid_num: int
    """Project number, e.g., 42."""

    title: str
    """Human-readable Project title."""

    description: str
    """Human-readable project markdown description."""

    project_dir: str
    """Project directory on disk. May include unresolved shell variables."""

    project_params_pdef: dict
    """Project-level job parameter default definitions."""

    owner_user_id: str
    """Object ID of user account that created this project."""

    created_at: str
    """Project creation date in ISO 8601 format."""

    deleted: bool
    """Whether this project has been deleted from the interface."""

    users_with_access: List[str]
    """Object IDs of user accounts that may access this project."""

    size: int
    """Computed size of project on disk."""

    last_accessed: ProjectLastAccessed
    """Details about when the project was last accessed by a user account."""

    archived: bool
    """Whether this project has been marked as archived from the inteface."""

    detached: bool
    """Whether this project is detached."""

    hidden: bool
    """Whether this project is hidden."""

    project_stats: dict
    """Computed project statistics."""

    generate_intermediate_results: bool
    """Whether intermediate results should be generated on this project."""


class JobDocument(TypedDict):
    """
    Specification for a Job document from the MongoDB database.
    """

    _id: str
    """MongoDB ID"""

    uid: str
    """Job unique ID, e.g., "J42"."""

    uid_num: int
    """Job number, e.g., 42."""

    project_uid: str
    """Project unique ID, e.g., "P3"."""

    project_uid_num: int
    """Project number, e.g., 3."""

    job_type: str
    """Job type identifier, e.g., "class2d"."""

    title: str
    """Human-readable job title."""

    description: str
    """Human-readable job markdown description."""

    status: JobStatus
    """Job scheduling status, e.g., "building", "queued", "running"."""

    created_at: str
    """Job creation date in ISO 8601 format."""

    created_by_user_id: Optional[str]
    """Object ID of user account that created this job."""

    deleted: bool
    """True if the job has been marked as deleted."""

    parents: List[str]
    """List of parent jobs UIDs based on input connections."""

    children: List[str]
    """List of child job UIDs based on output connections."""

    input_slot_groups: List[InputSlotGroup]
    """Input group specifications, including schema and connection information."""

    output_result_groups: List[OutputResultGroup]
    """Output group specifications."""

    output_results: List[OutputResult]
    """Aggregated output results specification (similar to
    ``output_result_groups`` with additional field information)."""

    params_base: Dict[str, Union[Param, EnumParam, PathParam]]
    """Job param specification and their base values. Each key represents a
    parameter name."""

    params_spec: Dict[str, ParamSpec]
    """User-specified parameter values. Each key is a parameter value. Not all
    keys from ``params_base`` are included here, only ones that were explicitly
    set."""

    workspace_uids: List[str]
    """List of workspace UIDs this job belongs to."""


class WorkspaceDocument(TypedDict):
    """
    Specification for a Workspace document from the MongoDB database.
    Live-related fields are not yet included.
    """

    _id: str
    """MongoDB ID"""

    uid: str
    """Workspace unique ID, e.g., "W1"."""

    uid_num: int
    """Workspace number, e.g., 1."""

    project_uid: str
    """Project unique ID, e.g., "P3"."""

    project_uid_num: int
    """Project number, e.g., 3."""

    created_at: str
    """Workspace creation date in ISO 8601 format."""

    created_by_user_id: str
    """Object ID of user account that created this workspace."""

    deleted: bool
    """True if the workspace has been marked as deleted."""

    title: str
    """Human-readable workspace title."""

    description: Optional[str]
    """Human-readable workspace markdown description."""

    workspace_type: Literal["base", "live"]
    """Either "live" or "base". """


class ResourceSlots(TypedDict):
    """
    Listings of available resources on a worker node that may be allocated for
    scheduling.
    """

    CPU: List[int]
    """List of available CPU core indices."""
    GPU: List[int]
    """List of available GPU indices."""
    RAM: List[int]
    """List of available 8GB slots."""


class FixedResourceSlots(TypedDict):
    """
    Available resource slots that only indicate presence, not the amount that
    may be allocated. (i.e., "SSD is available or not available")
    """

    SSD: bool
    """Whether this target thas an SSD"""


class Gpu(TypedDict):
    """
    GPU details for a target.
    """

    id: int
    """Index of GPU. Generally based on which PCI slot the GPU occupies."""
    name: str
    """Identifiable model name for this GPU, e.g.,"GeForce RTX 3090"."""
    mem: int
    """Amount of memory available on this GPU, in bytes."""


class SchedulerLane(TypedDict):
    """
    Description for a CryoSPARC scheduler lane.
    """

    name: str
    """Identifier for this lane."""
    type: Literal["node", "cluster"]
    """What kind of lane this is based on how on what kind of target(s) it contains."""
    title: str
    """Human-readable lane title."""
    desc: str
    """Human-readable lane description."""


class BaseSchedulerTarget(TypedDict):
    """
    Properties shared by both node and cluster scheduler targets.
    """

    lane: str
    """Lane name this target belongs to."""

    name: str
    """Identifier for this target."""

    title: str
    """Human-readable title for this target."""

    desc: Optional[str]
    """Human-readable description for this target."""

    hostname: str
    """Network machine hostname (same as name for for clusters)."""

    worker_bin_path: str
    """Path to cryosparc_worker/bin/cryosparcw executable."""

    cache_path: Optional[str]
    """Path the SSD cache scratch directory, if applicable."""

    cache_reserve_mb: int  # 10G default
    """Ensure at least this much space is free on the SSD scratch drive before
    caching."""

    cache_quota_mb: int
    """Do not cache more than this amoun on the SSD scrath drive.."""


class SchedulerTargetNode(BaseSchedulerTarget):
    """
    node-type scheduler target that does not include GPUs. Inherits from
    BaseSchedulerTarget_.

    .. _BaseSchedulerTarget:
        #cryosparc.spec.BaseSchedulerTarget
    """

    type: Literal["node"]
    """Node scheduler targets have type "node"."""

    ssh_str: str
    """Shell command used to access this node, e.g., ``ssh cryosparcuser@worker``."""

    resource_slots: ResourceSlots
    """Available compute resources."""

    resource_fixed: FixedResourceSlots
    """Available fixed resources."""

    monitor_port: Optional[int]
    """Not used."""


class SchedulerTargetGpuNode(SchedulerTargetNode):
    """
    node-type scheduler target that includes GPUs. Inherits from
    BaseSchedulerTarget_ and SchedulerTargetNode_.

    .. _BaseSchedulerTarget:
        #cryosparc.spec.BaseSchedulerTarget
    .. _SchedulerTargetNode:
        #cryosparc.spec.SchedulerTargetNode
    """

    gpus: List[Gpu]
    """Details about GPUs available on this node."""


class SchedulerTargetCluster(BaseSchedulerTarget):
    """
    Cluster-type scheduler targets. Inherits from BaseSchedulerTarget_.

    .. _BaseSchedulerTarget:
        #cryosparc.spec.BaseSchedulerTarget
    """

    type: Literal["cluster"]
    """Cluster scheduler targets have type "cluster"."""

    script_tpl: str
    """Full cluster submission script Jinja template."""

    send_cmd_tpl: str
    """Template command to access the cluster and running commands."""

    qsub_cmd_tpl: str
    """Template command to submit jobs to the cluster."""

    qstat_cmd_tpl: str
    """Template command to check the cluster job by its ID."""

    qdel_cmd_tpl: str
    """Template command to delete cluster jobs."""

    qinfo_cmd_tpl: str
    """Template command to check cluster queue info."""


SchedulerTarget = Union[SchedulerTargetNode, SchedulerTargetGpuNode, SchedulerTargetCluster]
"""
Scheduler target details.
"""


class JobSection(TypedDict):
    """
    Specification of available job types of a certain category.

    Example:

        >>> {
        ...     "name": "refinement",
        ...     "title": "3D Refinement",
        ...     "description: "...",
        ...     "contains" : [
        ...         "homo_refine",
        ...         "hetero_refine",
        ...         "nonuniform_refine",
        ...         "homo_reconstruct"
        ...     ]
        ... }
    """

    name: str
    """Section identifier."""
    title: str
    """Human-readable section title."""
    description: str
    """Human-readable section description."""
    contains: List[str]
    """Job type identifiers contained by this section."""


class MongoController(ABC, Generic[D]):
    """
    Abstract base class for Project, Workspace, Job classes and any other types
    that have underlying Mongo database documents.

    Generic type argument D is a typed dictionary definition for a Mongo
    document.

    :meta private:
    """

    _doc: Optional[D] = None

    @property
    def doc(self) -> D:
        if not self._doc:
            self.refresh()
        assert self._doc, "Could not refresh database document"
        return self._doc

    @abstractmethod
    def refresh(self):
        # Must be implemented in subclasses
        return self



from contextlib import contextmanager
from pathlib import PurePath
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    AsyncIterator,
    BinaryIO,
    Callable,
    ContextManager,
    Dict,
    Generator,
    Generic,
    Iterator,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)
from typing_extensions import Literal
import numpy as n

if TYPE_CHECKING:
    from numpy.typing import NDArray  # type: ignore

from .dtype import Shape

OpenTextMode = Literal["r", "w", "x", "a", "r+", "w+", "x+", "a+"]
"""
Text file read or write open modes.
"""

OpenBinaryMode = Literal["rb", "wb", "xb", "ab", "r+b", "w+b", "x+b", "a+b"]
"""
Binary file read or write open modes.
"""


T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")
E = TypeVar("E")  # error
INT = TypeVar("INT", bound=n.integer, covariant=True)

Result = Union[Tuple[T, Literal[None]], Tuple[Literal[None], E]]
"""
Use as the return type for functions that may return either a value or an
error.

Example:

    >>> def safe_divide(x: float, y: float) -> Result[int, str]:
    ...     if y == 0:
    ...         return None, "divide by zero detected"
    ...     else:
    ...         return x / y, None
"""


class hashcache(Dict[K, V], Generic[K, V]):
    """
    Simple utility class to cache the result of a mapping and avoid excessive
    heap allocation. Initialize with ``cache = hashcache(f)`` and use ``cache``
    as the mapping function.

    Examples:

        Unoptimized code for convering ``bytes`` to ``str``:

        >>> a = [b"Hello", b"Hello", b"Hello"]
        >>> strs = list(map(bytes.decode, a))

        The input array ``a`` has duplicate items. Using ``bytes.decode``
        directly in the mapping causes Python to allocate a fresh string for
        each bytes.

        Here is the optimized version with ``hashcache``:

        >>> a = [b"Hello", b"Hello", b"Hello"]
        >>> strs = list(map(hashcache(bytes.decode), a))

        This only allocates heap memory once for each unique item in the input
        list. After the first encounter, the map callable returns the
        previously-computed value of the same input, reducing heap usage and
        allocation calls by a factor of 3.

        For this to be most effective, ensure the given ``f`` function is pure
        and stable (i.e., always returns the same result for a given input).

        May also be used as a decorator (must take a single hashable argument):

        >>> @hashcache
        >>> def f(x): ...
    """

    __slots__ = ("factory",)
    __call__ = dict.__getitem__

    def __new__(cls, _f: Callable[[K], V]):
        return super().__new__(cls)

    def __init__(self, key_value_factory: Callable[[K], V]):
        super().__init__(self)
        self.factory = key_value_factory

    def __missing__(self, key):
        new = self.factory(key)
        self.__setitem__(key, new)
        return new


class BinaryIteratorIO(BinaryIO):
    """
    Read through a iterator that yields bytes as if it was a file.

    Args:
        iter (Iterator[bytes]): Generator that yields bytes
    """

    def __init__(self, iter: Iterator[bytes]):
        self._iter = iter
        self._left = b""

    def readable(self):
        """
        Returns:
            bool: True
        """
        return True

    def seekable(self) -> bool:
        """
        Returns:
            bool: False
        """
        return False

    def _read1(self, n: Optional[int] = None):
        while not self._left:
            try:
                self._left = next(self._iter)
            except StopIteration:
                break
        ret = self._left[:n]
        self._left = self._left[len(ret) :]
        return ret

    def read(self, n: Optional[int] = None) -> bytes:
        """
        Read the given number of bytes from the iterator

        Args:
            n (int, optional): Number of bytes to read. Reads all bytes if not
                specified. Defaults to None.

        Returns:
            bytes: Zero or more bytes read
        """
        r = []
        if n is None or n < 0:
            while True:
                m = self._read1()
                if not m:
                    break
                r.append(m)
        else:
            while n > 0:
                m = self._read1(n)
                if not m:
                    break
                n -= len(m)
                r.append(m)
        return b"".join(r)


class AsyncBinaryIteratorIO:
    """
    Similar to ``BinaryIteratorIO`` except the iterator yields bytes
    asynchronously.

    Args:
        iter (AsyncIterator[bytes]): Generator that asynchronously yields bytes.

    """

    def __init__(self, iter: Union[AsyncIterator[bytes], AsyncGenerator[bytes, None]]):
        self._iter = iter
        self._left = b""

    def readable(self):
        """
        Returns:
            bool: True
        """

        return True

    def seekable(self) -> bool:
        """
        Returns:
            bool: False
        """
        return False

    async def _read1(self, n: Optional[int] = None):
        while not self._left:
            try:
                self._left = await self._iter.__anext__()
            except StopAsyncIteration:
                break
        ret = self._left[:n]
        self._left = self._left[len(ret) :]
        return ret

    async def read(self, n: Optional[int] = None):
        """
        Read the given number of bytes from the async iterator.

        Args:
            n (int, optional): Number of bytes to read. Reads all bytes if not
                specified. Defaults to None.

        Returns:
            bytes: Zero or more bytes read
        """
        r = []
        if n is None or n < 0:
            while True:
                m = await self._read1()
                if not m:
                    break
                r.append(m)
        else:
            while n > 0:
                m = await self._read1(n)
                if not m:
                    break
                n -= len(m)
                r.append(m)
        return b"".join(r)


def first(it: Union[Iterator[V], Sequence[V]]) -> Optional[V]:
    """
    Get the first item from the given iterator. Returns None if the iterator is
    empty.

    Args:
        it (Iterator[V] | Sequence[V]): Iterator or list-like accessor.

    Returns:
        V | None: First item in the list or iterator (consuming it if required).
    """
    try:
        return it[0] if isinstance(it, Sequence) else next(it)
    except (StopIteration, IndexError):
        return None


def u32bytesle(x: int) -> bytes:
    """
    Get the uint32 bytes of for integer x in little endian.

    Args:
        x (int): Integer to encode.

    Returns:
        bytes: Encoded integer bytes
    """
    return x.to_bytes(4, "little", signed=False)


def u32intle(buffer: bytes) -> int:
    """
    Get int from buffer representing a uint32 integer in little endian.

    Args:
        buffer (bytes): 4 bytes representing little-endian integer.

    Returns:
        int: decoded integer
    """
    return int.from_bytes(buffer, "little", signed=False)


def strbytelen(s: str) -> int:
    """
    Get the number of bytes in a string's UTF-8 representation.

    Args:
        s (str): string to test encoding for.

    Returns:
        int: number of bytes in encoded string.
    """
    return len(str.encode(s))


def strencodenull(s: Any) -> bytes:
    """
    Encode string-like value into UTF-8 binary ending with a null-character
    terminator \0.

    Args:
        s (any): string-like entity to encode.

    Returns:
        bytes: encoded string
    """
    return ("" if s is None else str(s)).encode() + b"\0"


@contextmanager
def topen(file: Union[str, PurePath, IO[str]], mode: OpenTextMode = "r"):
    """
    "with open(...)" alias for text files that tranparently yields open file or
    file-like object.

    Args:
        file (str | Path | IO): Readable text file path or handle.
        mode (OpenTextMode, optional): File open mode. Defaults to "r".

    Yields:
        IO: Text file handle.
    """
    if isinstance(file, (str, PurePath)):
        with open(file, mode) as f:
            yield f
    else:
        yield file


@contextmanager
def bopen(file: Union[str, PurePath, IO[bytes]], mode: OpenBinaryMode = "rb"):
    """
    "with open(...)" alias for binary files that tranparently yields an open
    file or file-like object.

    Args:
        file (str | Path | IO): binary file path or handle
        mode (OpenBinaryMode, optional): File open mode. Defaults to "rb".

    Yields:
        IO: Binary file handle
    """
    if isinstance(file, (str, PurePath)):
        with open(file, mode) as f:
            yield f
    else:
        yield file


@overload
def noopcontext() -> ContextManager[None]:
    ...


@overload
def noopcontext(x: T) -> ContextManager[T]:
    ...


@contextmanager
def noopcontext(x: Optional[T] = None) -> Generator[Optional[T], None, None]:
    """
    Context manager that yields the given argument without modification.

    Args:
        x (T, optional): Anything. Defaults to None.

    Yields:
        T: the given argument
    """
    yield x


def padarray(arr: "NDArray", dim: Optional[int] = None, val: n.number = n.float32(0)):
    """
    Pad the given 2D or 3D array so that the x and y dimensions are equal to the
    given dimension. If not dimension is given, will use the maximum of the
    width and height.

    Args:
        arr (NDArray): 2D or 3D Numpy array to pad
        dim (int, optional): Minimum desired dimension of micrograph. Defaults
            to None.
        val (float, optional): Value to pad with, should be a numpy number
            instance with the same dtype as the given array. Defaults to
            n.float32(0).

    Returns:
        NDArray: Padded copy of given array.
    """
    arr = n.reshape(arr, (-1,) + arr.shape[-2:])
    nz, ny, nx = arr.shape
    idim = max(ny, nx) if dim is None else dim
    res = n.full((nz, idim, idim), val, dtype=arr.dtype)
    nya, nxa = ny // 2, nx // 2
    nyb, nxb = ny - nya, nx - nxa
    ya, yb = (idim // 2) - nya, (idim // 2) + nyb
    xa, xb = (idim // 2) - nxa, (idim // 2) + nxb
    res[:, ya:yb, xa:xb] = arr

    return n.reshape(res, res.shape[-2:]) if nz == 1 else res


def trimarray(arr: "NDArray", shape: Shape):
    """
    Crop the given 2D or 3D array into the given shape. Will trim from the
    middle. May be used to undo padding from ``padarray()`` function.

    Args:
        arr (NDArray): 3D or 3D numpy array to crop
        shape (Shape): New desired shape.

    Returns:
        NDArray: Trimmed copy of the given array.
    """
    assert len(shape) == 2, f"Invalid trim shape {shape}; must be 2D"
    arr = n.reshape(arr, (-1,) + arr.shape[-2:])
    z, x, y = arr.shape
    ny, nx = shape
    nya, nxa = ny // 2, nx // 2
    nyb, nxb = ny - nya, nx - nxa
    ya, yb = (x // 2) - nya, (x // 2) + nyb
    xa, xb = (y // 2) - nxa, (y // 2) + nxb
    res = arr[:, ya:yb, xa:xb]
    return n.reshape(res, res.shape[-2:]) if z == 1 else res


def default_rng(seed=None) -> "n.random.Generator":
    """
    Create a numpy random number generator or RandomState (depending on numpy
    version)

    Args:
        seed (Any, optional): Seed to initialize generator with. Defaults to None.

    Returns:
        numpy.random.Generator: Random number generator
    """
    try:
        return n.random.default_rng(seed)
    except AttributeError:
        return n.random.RandomState(seed)  # type: ignore


def random_integers(
    rng: "n.random.Generator",
    low: int,
    high: Optional[int] = None,
    size: Union[int, Shape, Literal[None]] = None,
    dtype: Type[INT] = n.uint64,
) -> "NDArray[INT]":
    """
    Generic way to get random integers from a numpy random generator (or
    RandomState for older numpy).

    Args:
        rng (numpy.random.Generator): random number generator
        low (int): Low integer value, inclusive
        high (int, optional): High integer value, exclusive. Defaults to None.
        size (Shape, optional): Size or shape of resulting array. Defaults to None.
        dtype (Type[numpy.integer], optional): Type of integer values to generate. Defaults to numpy.uint64.

    Returns:
        NDArray: Numpy array of randomly-generated integers.
    """
    try:
        f = rng.integers
    except AttributeError:
        f = rng.randint  # type: ignore
    return f(low=low, high=high, size=size, dtype=dtype)  # type: ignore



"""
Utilities and type definitions for working with dataset fields and column types.
"""
from typing import TYPE_CHECKING, Dict, List, Tuple, Type, Union
import json
from typing_extensions import Literal, TypedDict
import numpy as n

from .core import Data, DsetType

if TYPE_CHECKING:
    from numpy.typing import NDArray, DTypeLike

Shape = Tuple[int, ...]
"""A numpy shape tuple from ndarray.shape"""

DType = Union[str, Tuple[str, Shape]]
"""

    Can just be a single string such as "f4", "3u4" or "O".
    A datatype description of a ndarray entry.

    Can also be the a tuple with a string datatype name and its shape. For
    example, the following dtypes are equivalent.

    - "3u4"
    - "<u4", (3,))
"""

Field = Union[Tuple[str, str], Tuple[str, str, Shape]]
"""
    Description of a column in a numpy array with named fields

    Examples:
    - ("uid", "u8")
    - ("coords", "3f4")
    - ("coords", "<f4", (3,))
"""


class DatasetHeader(TypedDict):
    """
    Dataset header description when saving in CSDAT format.
    """

    dtype: List[Field]
    compression: Literal["snap"]
    compressed_fields: List[str]


DSET_TO_TYPE_MAP: Dict[DsetType, Type] = {
    DsetType.T_F32: n.float32,
    DsetType.T_F64: n.float64,
    DsetType.T_C32: n.complex64,
    DsetType.T_C64: n.complex128,
    DsetType.T_I8: n.int8,
    DsetType.T_I16: n.int16,
    DsetType.T_I32: n.int32,
    DsetType.T_I64: n.int64,
    DsetType.T_U8: n.uint8,
    DsetType.T_U16: n.uint16,
    DsetType.T_U32: n.uint32,
    DsetType.T_STR: n.uint64,  # Note: Prefer T_OBJ when working in Python
    DsetType.T_U64: n.uint64,
    DsetType.T_OBJ: n.object_,
}

TYPE_TO_DSET_MAP = {
    **{v: k for k, v in DSET_TO_TYPE_MAP.items()},
    **{
        float: DsetType.T_F64,
        complex: DsetType.T_C64,
        int: DsetType.T_I64,
        str: DsetType.T_OBJ,
        object: DsetType.T_OBJ,
    },
}

# Set of dataset fields that should not be compressed when saving in
# NEWEST_FORMAT
NEVER_COMPRESS_FIELDS = {"uid"}


def makefield(name: str, dtype: "DTypeLike") -> Field:
    dt = n.dtype(dtype)
    return (name, dt.base.str, dt.shape) if dt.shape else (name, dt.str)


def safe_makefield(name: str, dtype: "DTypeLike") -> Field:
    return ("uid", n.dtype(n.uint64).str) if name == "uid" else makefield(name, dtype)


def fielddtype(field: Field) -> DType:
    _, dt, *rest = field
    return (dt, rest[0]) if rest else dt


def arraydtype(a: "NDArray") -> DType:
    assert len(a.dtype.descr) == 1, "Cannot get dtype from record array"
    return (a.dtype.str, a.shape[1:]) if len(a.shape) > 1 else a.dtype.str


def dtypestr(dtype: "DTypeLike") -> str:
    dt = n.dtype(dtype)
    if dt.shape:
        shape = ",".join(map(str, dt.shape))
        return f"{dt.base.str[0]}{shape}{dt.base.str[1:]}"
    else:
        return dt.str


def get_data_field(data: Data, field: str) -> Field:
    return makefield(field, get_data_field_dtype(data, field))


def get_data_field_dtype(data: Data, field: str) -> "DTypeLike":
    t = data.type(field)
    if t == 0 or t not in DSET_TO_TYPE_MAP:
        raise KeyError(f"Unknown dataset field {field} or field type {t}")
    dt = n.dtype(DSET_TO_TYPE_MAP[t])
    shape = data.getshp(field)
    return (dt.str, shape) if shape else dt.str


def encode_dataset_header(fields: DatasetHeader) -> bytes:
    return json.dumps(fields).encode()


def decode_dataset_header(data: Union[bytes, dict]) -> DatasetHeader:
    try:
        header = json.loads(data) if isinstance(data, bytes) else data
        assert isinstance(header, dict), f"Incorrect decoded header type (expected dict, got {type(header)})"
        assert "dtype" in header and isinstance(
            header["dtype"], list
        ), 'Dataset header "dtype" key missing or has incorrect type'
        assert (
            "compression" in header and header["compression"] == "snap"
        ), 'Dataset header "compression" key missing or has incorrect type'
        assert (
            "compressed_fields" and header or isinstance(header["compressed_fields"], list)
        ), 'Dataset header "compressed_fields" key missing or has incorrect type'

        dtype: List[Field] = [(f, d, tuple(rest[0])) if rest else (f, d) for f, d, *rest in header["dtype"]]
        compression: Literal["snap"] = header["compression"]
        compressed_fields: List[str] = header["compressed_fields"]

        return DatasetHeader(dtype=dtype, compression=compression, compressed_fields=compressed_fields)
    except Exception as e:
        raise ValueError(f"Incorrect dataset field format: {data.decode() if isinstance(data, bytes) else data}") from e



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

from .dataset import Dataset
from .row import R
from .job import Job, ExternalJob
from .spec import MongoController, Datafield, Datatype, WorkspaceDocument

if TYPE_CHECKING:
    from .tools import CryoSPARC


class Workspace(MongoController[WorkspaceDocument]):
    """
    Accessor class to a workspace in CryoSPARC with ability create jobs and save
    results. Should be instantiated through `CryoSPARC.find_workspace`_ or
    `Project.find_workspace`_.

    .. _CryoSPARC.find_workspace:
        tools.html#cryosparc.tools.CryoSPARC.find_workspace

    .. _Project.find_workspace:
        project.html#cryosparc.project.Project.find_workspace
    """

    def __init__(self, cs: "CryoSPARC", project_uid: str, uid: str) -> None:
        self.cs = cs
        self.project_uid = project_uid
        self.uid = uid

    def refresh(self):
        """
        Reload this workspace from the CryoSPARC database.

        Returns:
            Workspace: self
        """
        self._doc = self.cs.cli.get_workspace(self.project_uid, self.uid)  # type: ignore
        return self

    def create_job(
        self,
        type: str,
        connections: Dict[str, Tuple[str, str]] = {},
        params: Dict[str, Any] = {},
        title: Optional[str] = None,
        desc: Optional[str] = None,
    ) -> Job:
        """
        Create a new job with the given type. Use the
        `CryoSPARC.get_job_sections`_ method to query available job types on
        the connected CryoSPARC instance.

        Args:
            project_uid (str): Project UID to create job in, e.g., "P3"
            workspace_uid (str): Workspace UID to create job in, e.g., "W1"
            type (str): Job type identifier, e.g., "homo_abinit"
            connections (dict[str, tuple[str, str]]): Initial input connections.
                Each key is an input name and each value is a (job uid, output
                name) tuple. Defaults to {}
            params (dict[str, any], optional): Specify parameter values.
                Defaults to {}.
            title (str, optional): Job title. Defaults to None.
            desc (str, optional): Job markdown description. Defaults to None.

        Returns:
            Job: created job instance. Raises error if job cannot be created.

        Examples:

            Create an Import Movies job.

            >>> from cryosparc.tools import CryoSPARC
            >>> cs = CryoSPARC()
            >>> workspace = cs.find_workspace("P3", "W3")
            >>> import_job = workspace.create_job("W1", "import_movies")
            >>> import_job.set_param("blob_paths", "/bulk/data/t20s/*.tif")
            True

            Create a 3-class ab-initio job connected to existing particles.

            >>> abinit_job = workspace.create_job("homo_abinit"
            ...     connections={"particles": ("J20", "particles_selected")}
            ...     params={"abinit_K": 3}
            ... )

        .. _CryoSPARC.get_job_sections:
            tools.html#cryosparc.tools.CryoSPARC.get_job_sections
        """
        return self.cs.create_job(
            self.project_uid, self.uid, type, connections=connections, params=params, title=title, desc=desc
        )

    def create_external_job(
        self,
        title: Optional[str] = None,
        desc: Optional[str] = None,
    ) -> ExternalJob:
        """
        Add a new External job to this workspace to save generated outputs to.

        Args:
            workspace_uid (str): Workspace UID to create job in, e.g., "W1"
            title (str, optional): Title for external job (recommended).
                Defaults to None.
            desc (str, optional): Markdown description for external job.
                Defaults to None.

        Returns:
            ExternalJob: created external job instance
        """
        return self.cs.create_external_job(self.project_uid, self.uid, title, desc)

    def save_external_result(
        self,
        dataset: Dataset[R],
        type: Datatype,
        name: Optional[str] = None,
        slots: Optional[List[Union[str, Datafield]]] = None,
        passthrough: Optional[Tuple[str, str]] = None,
        title: Optional[str] = None,
        desc: Optional[str] = None,
    ) -> str:
        """
        Save the given result dataset to a workspace.

        Args:
            dataset (Dataset): Result dataset.
            type (Datatype): Type of output dataset.
            name (str, optional): Name of output on created External job. Same
                as type if unspecified. Defaults to None.
            slots (list[str | Datafield], optional): List of slots expected to
                be created for this output such as ``location`` or ``blob``. Do
                not specify any slots that were passed through from an input
                unless those slots are modified in the output. Defaults to None.
            passthrough (tuple[str, str], optional): Indicates that this output
                inherits slots from the specified output. e.g., ``("J1",
                "particles")``. Defaults to None.
            title (str, optional): Human-readable title for this output.
                Defaults to None.
            desc (str, optional): Markdown description for this output. Defaults
                to None.

        Returns:
            str: UID of created job where this output was saved.

        Examples:

            Save all particle data

            >>> particles = Dataset()
            >>> workspace.save_external_result(particles, 'particle')
            "J43"

            Save new particle locations that inherit passthrough slots from a
            parent job

            >>> particles = Dataset()
            >>> workspace.save_external_result(
            ...     dataset=particles,
            ...     type='particle',
            ...     name='particles',
            ...     slots=['location'],
            ...     passthrough=('J42', 'selected_particles'),
            ...     title='Re-centered particles'
            ... )
            "J44"

            Save a result with multiple slots of the same type.

            >>> workspace.save_external_result(
            ...     dataset=particles,
            ...     type="particle",
            ...     name="particle_alignments",
            ...     slots=[
            ...         {"dtype": "alignments3D", "prefix": "alignments_class_0", "required": True},
            ...         {"dtype": "alignments3D", "prefix": "alignments_class_1", "required": True},
            ...         {"dtype": "alignments3D", "prefix": "alignments_class_2", "required": True},
            ...     ]
            ... )
            "J45"
        """
        return self.cs.save_external_result(
            self.project_uid,
            self.uid,
            dataset=dataset,
            type=type,
            name=name,
            slots=slots,
            passthrough=passthrough,
            title=title,
            desc=desc,
        )



from typing import Optional
import numpy as n

from .core import Data
from .dtype import Field, fielddtype
from .util import hashcache, strencodenull


class Column(n.ndarray):
    """
    Dataset column that inherits from the native numpy array interface.

    Note:
        Storing a column instance outside of a dataset prevents the whole
        dataset from getting garbage collected. Create a copy of its contents
        to prevent this, e.g., ``np.array(dset['ctf/phase_shift_rad']))``.

    Note:
        If new fields are added to the original dataset, a column instance may
        no longer be valid and must be retrieved again from
        ``dataset[fieldname]``.

    Examples:

        Storing column instances

        >>> dset = Dataset.allocate(1000, [('col1', 'f4'), ('col2', 'f4'), ('col3', 'f4')])
        Dataset(...)
        >>> col = dset['col1']
        >>> assert isinstance(col, Column)  # ok
        >>> del dset             # col still available but dset not garbage collected
        >>> col = np.array(col)  # dset now may be garbage collected

        Invalid column instances after adding columns

        >>> dset = Dataset.allocate(1000, [('col1', 'f4')])
        Dataset(...)
        >>> col = dset['col1']
        >>> dset.add_fields([('col2', 'f4'), ('col3', 'f4')])
        >>> np.sum(col)  # DANGER!! May result in invalid access
        >>> col = dset['col1']
        >>> np.sum(col)  # Retrieved from dataset, now valid
        0

    """

    __slots__ = ("_dataset",)
    _dataset: Optional[Data]

    def __new__(cls, field: Field, data: Data):
        dtype = n.dtype(fielddtype(field))
        nrow = data.nrow()
        shape = (nrow, *dtype.shape)
        buffer = data.getbuf(field[0]).memview if nrow else None
        obj = super().__new__(cls, shape=shape, dtype=dtype.base, buffer=buffer)  # type: ignore

        # Keep a reference to the data so that it only gets cleaned up when all
        # columns are cleaned up. No need to transfer this data during
        # __array_finalize__ because this NDColumn instance will be kept as
        # the base
        obj._dataset = data

        return obj

    def __array_wrap__(self, obj, **kwargs):
        # This prevents wrapping single results such as aggregations from n.sum
        # or n.median
        return obj[()] if obj.shape == () else super().__array_wrap__(obj, **kwargs)

    def to_fixed(self) -> "Column":
        """
        If this Column is composed of Python objects, convert to fixed-size
        strings. Otherwise the array is already in fixed form and may be
        returned as is.

        Returns:
            Column: Either same column with optionally converted data. If the
                data type is ``numpy.object_``, converts to ``numpy.bytes_``
                with a fixed element size based on the longest available string
                length.
        """
        if self.dtype.char == "O":
            return n.vectorize(hashcache(strencodenull), otypes="S")(self)
        else:
            return self



"""
Main module exporting the ``CryoSPARC`` class for interfacing with a CryoSPARC
instance from Python

Examples:

    >>> from cryosparc.tools import CryoSPARC
    >>> license = "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx"
    >>> email = "ali@example.com"
    >>> password = "password123"
    >>> cs = CryoSPARC(
    ...     license=license,
    ...     email=email,
    ...     password=password,
    ...     host="localhost",
    ...     base_port=39000
    ... )
    >>> project = cs.find_project("P3")

"""
from io import BytesIO
from pathlib import Path, PurePath, PurePosixPath
from typing import IO, TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Tuple, Union
import os
import re
import tempfile
import numpy as n

if TYPE_CHECKING:
    from numpy.typing import NDArray  # type: ignore

from . import mrc
from .command import CommandClient, make_json_request, make_request
from .dataset import DEFAULT_FORMAT, Dataset
from .row import R
from .project import Project
from .workspace import Workspace
from .job import ExternalJob, Job
from .spec import (
    ASSET_EXTENSIONS,
    AssetDetails,
    Datafield,
    Datatype,
    JobSection,
    SchedulerLane,
    SchedulerTarget,
)
from .util import bopen, noopcontext, padarray, trimarray


ONE_MIB = 2**20  # bytes in one mebibyte

LICENSE_REGEX = re.compile(r"[a-f\d]{8}-[a-f\d]{4}-[a-f\d]{4}-[a-f\d]{4}-[a-f\d]{12}")
"""Regular expression for matching CryoSPARC license IDs."""

SUPPORTED_EXPOSURE_FORMATS = {
    "MRC",
    "MRCS",
    "TIFF",
    "CMRCBZ2",
    "MRCBZ2",
    "EER",
}
"""
Supported micrograph file formats.
"""


class CryoSPARC:
    """
    High-level session class for interfacing with a CryoSPARC instance.

    Initialize with the host and base port of the running CryoSPARC instance.
    This host and (at minimum) ``base_port + 2``, ``base_port + 3`` and
    ``base_port + 5`` should be accessible on the network.

    Args:
        license (str, optional): CryoSPARC license key. Defaults to
            ``os.getenv("CRYOSPARC_LICENSE_ID")``.
        host (str, optional): Hostname or IP address running CryoSPARC master.
            Defaults to ``os.getenv("CRYOSPARC_MASTER_HOSTNAME", "localhost")``.
        base_port (int, optional): CryoSPARC services base port number. Defaults
            to ``os.getenv("CRYOSPARC_MASTER_HOSTNAME", 39000)``.
        email (str, optional): CryoSPARC user account email address. Defaults
            to ``os.getenv("CRYOSPARC_EMAIL")``.
        password (str, optional): CryoSPARC user account password address.
            Defaults to ``os.getenv("CRYOSPARC_PASSWORD")``.
        timeout (int, optional): Timeout error for HTTP requests to CryoSPARC
            command services. Defaults to 300.

    Attributes:
        cli (CommandClient): HTTP/JSONRPC client for ``command_core`` service (port + 2).
        vis (CommandClient): HTTP/JSONRPC client for ``command_vis`` service (port + 3).
        rtp (CommandClient): HTTP/JSONRPC client for ``command_rtp`` service (port + 5).
        user_id (str): Mongo object ID of user account performing operations for this session.

    Examples:

        Load project job and micrographs

        >>> from cryosparc.tools import CryoSPARC
        >>> cs = CryoSPARC(
        ...     license="xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
        ...     email="ali@example.com",
        ...     password="password123",
        ...     base_port=39000
        ... )
        >>> job = cs.find_job("P3", "J42")
        >>> micrographs = job.load_output('exposures')

        Remove corrupt exposures (assumes ``is_mic_corrupt`` function)

        >>> filtered_micrographs = micrographs.query(
        ...     lambda mic: is_mic_corrupt(mic["micrograph_blob/path"])
        ... )
        >>> cs.save_external_result(
        ...     project_uid="P3",
        ...     workspace_uid="W1",
        ...     dataset=filtered_micrographs,
        ...     type="exposure",
        ...     name="filtered_exposures",
        ...     passthrough=("J42", "exposures")
        ... )
        "J43"
    """

    cli: CommandClient
    vis: CommandClient
    rtp: CommandClient
    user_id: str  # session user ID

    def __init__(
        self,
        license: str = os.getenv("CRYOSPARC_LICENSE_ID", ""),
        host: str = os.getenv("CRYOSPARC_MASTER_HOSTNAME", "localhost"),
        base_port: int = int(os.getenv("CRYOSPARC_BASE_PORT", 39000)),
        email: str = os.getenv("CRYOSPARC_EMAIL", ""),
        password: str = os.getenv("CRYOSPARC_PASSWORD", ""),
        timeout: int = 300,
    ):
        assert LICENSE_REGEX.fullmatch(license), f"Invalid or unspecified CryoSPARC license ID {license}"
        assert email, "Invalid or unspecified email"
        assert password, "Invalid or unspecified password"

        self.cli = CommandClient(
            service="command_core", host=host, port=base_port + 2, headers={"License-ID": license}, timeout=timeout
        )
        self.vis = CommandClient(
            service="command_vis", host=host, port=base_port + 3, headers={"License-ID": license}, timeout=timeout
        )
        self.rtp = CommandClient(
            service="command_rtp", host=host, port=base_port + 5, headers={"License-ID": license}, timeout=timeout
        )
        try:
            self.user_id = self.cli.get_id_by_email_password(email, password)  # type: ignore
        except Exception as e:
            raise RuntimeError("Could not complete CryoSPARC authentication with given credentials") from e

    def test_connection(self):
        """
        Verify connection to CryoSPARC command services.

        Returns:
            bool: True if connection succeeded, False otherwise
        """
        if self.cli.test_connection():  # type: ignore
            print(f"Connection succeeded to CryoSPARC command_core at {self.cli._url}")
        else:
            print(f"Connection FAILED to CryoSPARC command_core at {self.cli._url}")
            return False

        with make_request(self.vis, method="get") as response:
            if response.read():
                print(f"Connection succeeded to CryoSPARC command_vis at {self.vis._url}")
            else:
                print(f"Connection FAILED to CryoSPARC command_vis at {self.vis._url}")
                return False

        with make_request(self.rtp, method="get") as response:
            if response.read():
                print(f"Connection succeeded to CryoSPARC command_rtp at {self.rtp._url}")
            else:
                print(f"Connection FAILED to CryoSPARC command_rtp at {self.rtp._url}")
                return False

        return True

    def get_lanes(self) -> List[SchedulerLane]:
        """
        Get a list of available scheduler lanes.

        Returns:
            list[SchedulerLane]: Details about available lanes.
        """
        return self.cli.get_scheduler_lanes()  # type: ignore

    def get_targets(self, lane: Optional[str] = None) -> List[SchedulerTarget]:
        """
        Get a list of available scheduler targets.

        Args:
            lane (str, optional): Only get targets from this specific lane.
                Returns all targets if not specified. Defaults to None.

        Returns:
            list[SchedulerTarget]: Details about available targets.
        """
        targets: List[SchedulerTarget] = self.cli.get_scheduler_targets()  # type: ignore
        if lane is not None:
            targets = [t for t in targets if t["lane"] == lane]
        return targets

    def get_job_sections(self) -> List[JobSection]:
        """
        Get a summary of job types available for this instance, organized by
        category.

        Returns:
            list[JobSection]: List of job section dictionaries. Job types
                are listed in the ``"contains"`` key in each dictionary.
        """
        return self.cli.get_job_sections()  # type: ignore

    def find_project(self, project_uid: str) -> Project:
        """
        Get a project by its unique ID.

        Args:
            project_uid (str): project unique ID, e.g., "P3"

        Returns:
            Project: project instance
        """
        project = Project(self, project_uid)
        project.refresh()
        return project

    def find_workspace(self, project_uid: str, workspace_uid: str) -> Workspace:
        """
        Get a workspace accessor instance for the workspace in the given project
        with the given UID. Fails with an error if workspace does not exist.

        Args:
            project_uid (str): Project unique ID, e.g,. "P3"
            workspace_uid (str): Workspace unique ID, e.g., "W1"

        Returns:
            Workspace: accessor instance
        """
        workspace = Workspace(self, project_uid, workspace_uid)
        return workspace.refresh()

    def find_job(self, project_uid: str, job_uid: str) -> Job:
        """
        Get a job by its unique project and job ID.

        Args:
            project_uid (str): project unique ID, e.g., "P3"
            job_uid (str): job unique ID, e.g., "J42"

        Returns:
            Job: job instance
        """
        job = Job(self, project_uid, job_uid)
        job.refresh()
        return job

    def find_external_job(self, project_uid: str, job_uid: str) -> ExternalJob:
        """
        Get the External job accessor instance for an External job in this
        project with the given UID. Fails if the job does not exist or is not an
        external job.

        Args:
            project_uid (str): Project unique ID, e.g,. "P3"
            job_uid (str): Job unique ID, e.g,. "J42"

        Raises:
            TypeError: If job is not an external job

        Returns:
            ExternalJob: accessor instance
        """
        job = ExternalJob(self, project_uid, job_uid)
        job.refresh()
        if job.doc["job_type"] != "snowflake":
            raise TypeError(f"Job {project_uid}-{job_uid} is not an external job")
        return job

    def create_workspace(self, project_uid: str, title: str, desc: Optional[str] = None) -> Workspace:
        """
        Create a new empty workspace in the given project.

        Args:
            project_uid (str): Project UID to create in, e.g., "P3".
            title (str): Title of new workspace.
            desc (str, optional): Markdown text description. Defaults to None.

        Returns:
            Workspace: created workspace instance
        """
        workspace_uid: str = self.cli.create_empty_workspace(  # type: ignore
            project_uid=project_uid, created_by_user_id=self.user_id, title=title, desc=desc
        )
        return self.find_workspace(project_uid, workspace_uid)

    def create_job(
        self,
        project_uid: str,
        workspace_uid: str,
        type: str,
        connections: Dict[str, Tuple[str, str]] = {},
        params: Dict[str, Any] = {},
        title: Optional[str] = None,
        desc: Optional[str] = None,
    ) -> Job:
        """
        Create a new job with the given type. Use `CryoSPARC.get_job_sections`_
        to query available job types on the connected CryoSPARC instance.

        Args:
            project_uid (str): Project UID to create job in, e.g., "P3"
            workspace_uid (str): Workspace UID to create job in, e.g., "W1"
            type (str): Job type identifier, e.g., "homo_abinit"
            connections (dict[str, tuple[str, str]]): Initial input connections.
                Each key is an input name and each value is a (job uid, output
                name) tuple. Defaults to {}
            params (dict[str, any], optional): Specify parameter values.
                Defaults to {}.
            title (str, optional): Job title. Defaults to None.
            desc (str, optional): Job markdown description. Defaults to None.

        Returns:
            Job: created job instance. Raises error if job cannot be created.

        Examples:

            Create an Import Movies job.

            >>> from cryosparc.tools import CryoSPARC
            >>> cs = CryoSPARC()
            >>> import_job = cs.create_job("P3", "W1", "import_movies")
            >>> import_job.set_param("blob_paths", "/bulk/data/t20s/*.tif")
            True

            Create a 3-class ab-initio job connected to existing particles.

            >>> abinit_job = cs.create_job("P3", "W1", "homo_abinit"
            ...     connections={"particles": ("J20", "particles_selected")}
            ...     params={"abinit_K": 3}
            ... )

        .. _CryoSPARC.get_job_sections:
            #cryosparc.tools.CryoSPARC.get_job_sections
        """
        job_uid: str = self.cli.create_new_job(  # type: ignore
            job_type=type, project_uid=project_uid, workspace_uid=workspace_uid, title=title, desc=desc
        )
        job = self.find_job(project_uid, job_uid)
        for input_name, (parent_job, output_name) in connections.items():
            job.connect(parent_job, output_name, input_name, refresh=False)
        for k, v in params.items():
            job.set_param(k, v, refresh=False)
        if connections or params:
            job.refresh()
        return job

    def create_external_job(
        self,
        project_uid: str,
        workspace_uid: str,
        title: Optional[str] = None,
        desc: Optional[str] = None,
    ) -> ExternalJob:
        """
        Add a new External job to this project to save generated outputs to.

        Args:
            project_uid (str): Project UID to create in, e.g., "P3"
            workspace_uid (str): Workspace UID to create job in, e.g., "W1"
            title (str, optional): Title for external job (recommended).
                Defaults to None.
            desc (str, optional): Markdown description for external job.
                Defaults to None.

        Returns:
            ExternalJob: created external job instance
        """
        job_uid: str = self.vis.create_external_job(  # type: ignore
            project_uid=project_uid, workspace_uid=workspace_uid, user=self.user_id, title=title, desc=desc
        )
        return self.find_external_job(project_uid, job_uid)

    def save_external_result(
        self,
        project_uid: str,
        workspace_uid: Optional[str],
        dataset: Dataset[R],
        type: Datatype,
        name: Optional[str] = None,
        slots: Optional[List[Union[str, Datafield]]] = None,
        passthrough: Optional[Tuple[str, str]] = None,
        title: Optional[str] = None,
        desc: Optional[str] = None,
    ) -> str:
        """
        Save the given result dataset to the project. Specify at least the
        dataset to save and the type of data.

        Returns UID of the External job where the results were saved.

        Examples:

            Save all particle data

            >>> particles = Dataset()
            >>> cs.save_external_result("P1", "W1", particles, 'particle')
            "J43"

            Save new particle locations that inherit passthrough slots from a
            parent job.

            >>> particles = Dataset()
            >>> cs.save_external_result(
            ...     project_uid="P1",
            ...     workspace_uid="W1",
            ...     dataset=particles,
            ...     type='particle',
            ...     name='particles',
            ...     slots=['location'],
            ...     passthrough=('J42', 'selected_particles'),
            ...     title='Re-centered particles'
            ... )
            "J44"

            Save a result with multiple slots of the same type.

            >>> cs.save_external_result(
            ...     project_uid="P1",
            ...     workspace_uid="P1",
            ...     dataset=particles,
            ...     type="particle",
            ...     name="particle_alignments",
            ...     slots=[
            ...         {"dtype": "alignments3D", "prefix": "alignments_class_0", "required": True},
            ...         {"dtype": "alignments3D", "prefix": "alignments_class_1", "required": True},
            ...         {"dtype": "alignments3D", "prefix": "alignments_class_2", "required": True},
            ...     ]
            ... )
            "J45"

        Args:
            project_uid (str): Project UID to save results into.
            workspace_uid (str | None): Workspace UID to save results into.
                Specify ``None`` to auto-select a workspace.
            dataset (Dataset): Result dataset.
            type (Datatype): Type of output dataset.
            name (str, optional): Name of output on created External job. Same
                as type if unspecified. Defaults to None.
            slots (list[str | Datafield], optional): List of slots expected to
                be created for this output such as ``location`` or ``blob``. Do
                not specify any slots that were passed through from an input
                unless those slots are modified in the output. Defaults to None.
            passthrough (tuple[str, str], optional): Indicates that this output
                inherits slots from the specified output. e.g.,
                ``("J1", "particles")``. Defaults to None.

            title (str, optional): Human-readable title for this output.
                Defaults to None.
            desc (str, optional): Markdown description for this output. Defaults
                to None.

        Returns:
            str: UID of created job where this output was saved
        """
        # Check slot names if present. If not provided, use all slots specified
        # in the dataset prefixes.
        prefixes = dataset.prefixes()
        if slots is None:
            slots = list(prefixes)
        slot_names = {s if isinstance(s, str) else s["prefix"] for s in slots}
        assert slot_names.intersection(prefixes) == slot_names, "Given dataset missing required slots"

        passthrough_str = ".".join(passthrough) if passthrough else None
        job_uid, output = self.vis.create_external_result(  # type: ignore
            project_uid=project_uid,
            workspace_uid=workspace_uid,
            type=type,
            name=name,
            slots=slots,
            passthrough=passthrough_str,
            user=self.user_id,
            title=title,
            desc=desc,
        )

        job = self.find_external_job(project_uid, job_uid)
        with job.run():
            job.save_output(output, dataset)

        return job.uid

    def list_files(
        self, project_uid: str, prefix: Union[str, PurePosixPath] = "", recursive: bool = False
    ) -> List[str]:
        """
        Get a list of files inside the project directory.

        Args:
            project_uid (str): Project unique ID, e.g., "P3"
            prefix (str | Path, optional): Subdirectory inside project to list.
                Defaults to "".
            recursive (bool, optional): If True, lists files recursively.
                Defaults to False.

        Returns:
            list[str]: List of file paths relative to the project directory.
        """
        return self.vis.list_project_files(  # type: ignore
            project_uid=project_uid,
            prefix=str(prefix),
            recursive=recursive,
        )

    def download(self, project_uid: str, path_rel: Union[str, PurePosixPath]):
        """
        Open a file in the given project for reading. Use to get files from a
        remote CryoSPARC instance whose the project directories are not
        available on the client file system.

        Args:
            project_uid (str): Short unique ID of CryoSPARC project, e.g., "P3"
            path_rel (str | Path): Relative path to file in project directory

        Yields:
            HTTPResponse: Use a context manager to read the file from the
            request body

        Examples:
            Download a job's metadata

            >>> cs = CryoSPARC()
            >>> with cs.download('P3', 'J42/job.json') as res:
            >>>     job_data = json.loads(res.read())

        """
        data = {"project_uid": project_uid, "path_rel": str(path_rel)}
        return make_json_request(self.vis, "/get_project_file", data=data)

    def download_file(
        self, project_uid: str, path_rel: Union[str, PurePosixPath], target: Union[str, PurePath, IO[bytes]]
    ):
        """
        Download a file from the project directory to the given writeable target.

        Args:
            project_uid (str): project unique ID, e.g., "P3"
            path_rel (str | Path): Relative path of file in project directory.
            target (str | Path | IO): Local file path, directory path or writeable
                file handle to write response data.

        Returns:
            Path | IO: resulting target path or file handle.
        """
        if isinstance(target, (str, PurePath)):
            target = Path(target)
            if target.is_dir():
                target /= PurePath(path_rel).name
        with bopen(target, "wb") as f:
            with self.download(project_uid, path_rel) as response:
                data = response.read(ONE_MIB)
                while data:
                    f.write(data)
                    data = response.read(ONE_MIB)
        return target

    def download_dataset(self, project_uid: str, path_rel: Union[str, PurePosixPath]):
        """
        Download a .cs dataset file from the given relative path in the project
        directory.

        Args:
            project_uid (str): project unique ID, e.g., "P3"
            path_rel (str | Path): Realtive path to .cs file in project
            directory.

        Returns:
            Dataset: Loaded dataset instance
        """
        with self.download(project_uid, path_rel) as response:
            size = response.headers.get("Content-Length")
            mime = response.headers.get("Content-Type")
            if mime == "application/x-cryosparc-dataset":
                # Stream format; can load directly without seek
                return Dataset.load(response)

            # Numpy format, cannot load directly because requires seekable
            if size and int(size) < ONE_MIB:
                # Smaller than 1MiB, just read all into memory and load
                return Dataset.load(BytesIO(response.read()))

            # Read into temporary file in 1MiB chunks. Load from that temporary file
            with tempfile.TemporaryFile("w+b", suffix=".cs") as f:
                data = response.read(ONE_MIB)
                while data:
                    f.write(data)
                    data = response.read(ONE_MIB)
                f.seek(0)
                return Dataset.load(f)

    def download_mrc(self, project_uid: str, path_rel: Union[str, PurePosixPath]):
        """
        Download a .mrc file from the given relative path in the project
        directory.

        Args:
            project_uid (str): project unique ID, e.g., "P3"
            path_rel (str | Path): Relative path to .mrc file in project
                directory.

        Returns:
            tuple[Header, NDArray]: MRC file header and data as a numpy array
        """
        with self.download(project_uid, path_rel) as response:
            with tempfile.TemporaryFile("w+b", suffix=".cs") as f:
                data = response.read(ONE_MIB)
                while data:
                    f.write(data)
                    data = response.read(ONE_MIB)
                f.seek(0)
                return mrc.read(f)  # FIXME: Optimize file reading

    def list_assets(self, project_uid: str, job_uid: str) -> List[AssetDetails]:
        """
        Get a list of files available in the database for given job. Returns a
        list with details about the assets. Each entry is a dict with a ``_id``
        key which may be used to download the file with the ``download_asset``
        method.

        Args:
            project_uid (str): project unique ID, e.g., "P3"
            job_uid (str): job unique ID, e.g., "J42"

        Returns:
            list[AssetDetails]: Asset details
        """
        return self.vis.list_job_files(project_uid=project_uid, job_uid=job_uid)  # type: ignore

    def download_asset(self, fileid: str, target: Union[str, PurePath, IO[bytes]]):
        """
        Download a file from CryoSPARC's MongoDB GridFS storage.

        Args:
            fileid (str): GridFS file object ID
            target (str | Path | IO): Local file path, directory path or
                writeable file handle to write response data.

        Returns:
            Path | IO: resulting target path or file handle.
        """
        with make_json_request(self.vis, url="/get_job_file", data={"fileid": fileid}) as response:
            if isinstance(target, (str, PurePath)):
                target = Path(target)
                if target.is_dir():
                    # Try to get download filename and content type from
                    # headers. If cannot be determined, defaults to "file.dat"
                    content_type: str = response.headers.get_content_type()
                    attachment_filename: Optional[str] = response.headers.get_filename()
                    target /= attachment_filename or f"file.{ASSET_EXTENSIONS.get(content_type, 'dat')}"
            with bopen(target, "wb") as f:
                data = response.read(ONE_MIB)
                while data:
                    f.write(data)
                    data = response.read(ONE_MIB)

            return target

    def upload(
        self, project_uid: str, target_path_rel: Union[str, PurePosixPath], source: Union[str, bytes, PurePath, IO]
    ):
        """
        Upload the given source file to the project directory at the given
        relative path.

        Args:
            project_uid (str): project unique ID, e.g., "P3"
            target_path_rel (str | Path): Relative target path in project
                directory.
            source (str | bytes | Path | IO): Local path or file handle to
                upload. May also specified as raw bytes.
        """
        with open(source, "rb") if isinstance(source, (str, PurePath)) else noopcontext(source) as f:
            url = f"/projects/{project_uid}/files"
            query = {"path": target_path_rel}
            with make_request(self.vis, url=url, query=query, data=f) as res:
                assert res.status >= 200 and res.status < 300, (
                    f"Could not upload project {project_uid} file {target_path_rel}.\n"
                    f"Response from CryoSPARC ({res.status}): {res.read().decode()}"
                )

    def upload_dataset(
        self, project_uid: str, target_path_rel: Union[str, PurePosixPath], dset: Dataset, format: int = DEFAULT_FORMAT
    ):
        """
        Upload a dataset as a CS file into the project directory.

        Args:
            project_uid (str): project unique ID, e.g., "P3"
            target_path_rel (str | Path): relative path to save dataset in
                project directory. Should have a ``.cs`` extension.
            dset (Dataset): dataset to save.
            format (int): format to save in from ``cryosparc.dataset.*_FORMAT``,
                defaults to NUMPY_FORMAT)
        """
        if len(dset) < 100:
            # Probably small enough to upload from memory
            f = BytesIO()
            dset.save(f, format=format)
            f.seek(0)
            return self.upload(project_uid, target_path_rel, f)

        # Write to temp file first
        with tempfile.TemporaryFile("w+b") as f:
            dset.save(f, format=format)
            f.seek(0)
            return self.upload(project_uid, target_path_rel, f)

    def upload_mrc(self, project_uid: str, target_path_rel: Union[str, PurePosixPath], data: "NDArray", psize: float):
        """
        Upload a numpy 2D or 3D array to the job directory as an MRC file.

        Args:
            project_uid (str): project unique ID, e.g., "P3"
            target_path_rel (str | Path): filename or relative path. Should have
                ``.mrc`` extension.
            data (NDArray): Numpy array with MRC file data.
            psize (float): Pixel size to include in MRC header.
        """
        with tempfile.TemporaryFile("w+b") as f:
            mrc.write(f, data, psize)
            f.seek(0)
            return self.upload(project_uid, target_path_rel, f)

    def mkdir(
        self,
        project_uid: str,
        target_path_rel: Union[str, PurePosixPath],
        parents: bool = False,
        exist_ok: bool = False,
    ):
        """
        Create a directory in the given project.

        Args:
            project_uid (str): Target project directory
            target_path_rel (str | Path): Relative path to create inside project
                directory.
            parents (bool, optional): If True, any missing parents are created
                as needed. Defaults to False.
            exist_ok (bool, optional): If True, does not raise an error for
                existing directories. Still raises if the target path is not a
                directory. Defaults to False.
        """
        self.vis.project_mkdir(  # type: ignore
            project_uid=project_uid,
            path_rel=str(target_path_rel),
            parents=parents,
            exist_ok=exist_ok,
        )

    def cp(
        self, project_uid: str, source_path_rel: Union[str, PurePosixPath], target_path_rel: Union[str, PurePosixPath]
    ):
        """
        Copy a file or folder within a project to another location within that
        same project. Note that argument order is reversed from
        equivalent ``cp`` command.

        Args:
            project_uid (str): Target project UID, e.g., "P3".
            source_path_rel (str | Path): Relative path in project of source
                file or folder to copy.
            target_path_rel (str | Path): Relative path in project to copy to.
        """
        self.vis.project_cp(  # type: ignore
            project_uid=project_uid,
            source_path_rel=str(source_path_rel),
            target_path_rel=str(target_path_rel),
        )

    def symlink(
        self, project_uid: str, source_path_rel: Union[str, PurePosixPath], target_path_rel: Union[str, PurePosixPath]
    ):
        """
        Create a symbolic link in the given project. May only create links for
        files within the project. Note that argument order is reversed from
        ``ln -s``.

        Args:
            project_uid (str): Target project UID, e.g., "P3".
            source_path_rel (str | Path): Relative path in project to file from
                which to create symlink.
            target_path_rel (str | Path): Relative path in project to new
                symlink.
        """
        self.vis.project_symlink(  # type: ignore
            project_uid=project_uid,
            source_path_rel=str(source_path_rel),
            target_path_rel=str(target_path_rel),
        )


def get_import_signatures(abs_paths: Union[str, Iterable[str], "NDArray"]):
    """
    Get list of import signatures for the given path or paths.

    Args:
        abs_paths (str | Iterable[str] | NDArray): Absolute path or list of file
            paths.

    Returns:
        list[int]: Import signatures as 64-bit numpy integers
    """
    from hashlib import sha1

    if isinstance(abs_paths, str):
        abs_paths = [abs_paths]

    return [n.frombuffer(sha1(path.encode()).digest()[:8], dtype=n.uint64)[0] for path in abs_paths]


def get_exposure_format(data_format: str, voxel_type: Optional[str] = None) -> str:
    """
    Get the ``movie_blob/format`` or ``micrograph_blob`` format value for an
    exposure type, where ``data_format`` is one of

    - "MRC"
    - "MRCS"
    - "TIFF"
    - "CMRCBZ2"
    - "MRCBZ2"
    - "EER"

    And ``voxel_type`` (if specified) is one of

    - "16 BIT FLOAT":
    - "32 BIT FLOAT"
    - "SIGNED 16 BIT INTEGER"
    - "UNSIGNED 8 BIT INTEGER"
    - "UNSIGNED 16 BIT INTEGER"

    Args:
        data_format (str): One of ``SUPPORTED_EXPOSURE_FORMATS`` such as
            ``"TIFF"`` or ``"MRC"``. The value of the ``<dataFormat>`` tag in an
            EPU XML file.
        voxel_type (str, optional): The value of the ``<voxelType>`` tag in an
            EPU file such as ``"32 BIT FLOAT"``. Required when ``data_format``
            is ``MRC`` or ``MRCS``. Defaults to None.

    Returns:
        str: The format string to save into the ``{prefix}/format`` field of a
            CryoSPARC exposure dataset. e.g., ``"TIFF"`` or ``"MRC/2"``
    """
    assert data_format in SUPPORTED_EXPOSURE_FORMATS, f"Unsupported exposure format {data_format}"
    if data_format not in {"MRC", "MRCS"}:
        return data_format

    assert (
        voxel_type and voxel_type in mrc.VOXEL_TYPES
    ), f'Unsupported voxel type "{voxel_type}" specified with MRC exposure format'
    return f"MRC/{mrc.VOXEL_TYPES[voxel_type]}"


def downsample(arr: "NDArray", factor: int = 2):
    """
    Downsample a micrograph or movie by the given factor.

    Args:
        arr (NDArray): 2D or 3D numpy array factor (int, optional): How much to
            reduce size by. e.g., a factor of 2 would reduce a 1024px MRC to
            512px, and a factor of 3 would reduce it to 256px. Defaults to 2.

    Returns:
        NDArray: Downsampled MRC file
    """
    assert factor >= 1, "Must bin by a factor of 1 or greater"
    arr = n.reshape(arr, (-1,) + arr.shape[-2:])
    nz, ny, nx = arr.shape
    clipx = (nx // factor) * factor
    clipy = (ny // factor) * factor
    shape = (nz, (clipy // factor), (clipx // factor)) if nz > 1 else ((clipy // factor), (clipx // factor))
    out = arr[:, :clipy, :clipx].reshape(nz, clipy, (clipx // factor), factor)
    out = out.sum(axis=-1)
    out = out.reshape(nz, (clipy // factor), factor, -1).sum(axis=-2)
    return out.reshape(shape)


def lowpass2(arr: "NDArray", psize_A: float, cutoff_resolution_A: float = 0.0, order: float = 1.0):
    """
    Apply butterworth lowpass filter to the 2D array data with the given
    pixel size (``psize_A``). ``cutoff_resolution_A`` should be a non-negative
    number specified in Angstroms.

    Args:
        arr (NDArray): 2D numpy array to apply lowpass to.
        psize_A (float): Pixel size of array data.
        cutoff_resolution_A (float, optional): Cutoff resolution, in Angstroms.
            Defaults to 0.0.
        order (float, optional): Filter order. Defaults to 1.0.

    Returns:
        NDArray: Lowpass-filtered copy of given numpy array
    """
    assert cutoff_resolution_A > 0, "Lowpass filter amount must be non-negative"
    assert len(arr.shape) == 2 or (len(arr.shape) == 3 and arr.shape[0] == 1), (
        f"Cannot apply low-pass filter on data with shape {arr.shape}; " "must be two-dimensional"
    )

    arr = n.reshape(arr, arr.shape[-2:])
    shape = arr.shape
    if arr.shape[0] != arr.shape[1]:
        arr = padarray(arr, val=n.mean(arr))

    radwn = (psize_A * arr.shape[-1]) / cutoff_resolution_A
    inverse_cutoff_wn2 = 1.0 / radwn**2

    farr = n.fft.rfft2(arr)
    ny, nx = farr.shape
    yp = 0

    for y in range(ny // 2):
        yp = (ny // 2) if y == 0 else ny - y

        # y goes from DC to one before nyquist
        # x goes from DC to one before nyquist
        r2 = (n.arange(nx - 1) ** 2) + (y * y)
        f = 1.0 / (1.0 + (r2 * inverse_cutoff_wn2) ** order)
        farr[y][:-1] *= f
        farr[yp][:-1] *= f

    # zero nyquist at the end
    farr[ny // 2] = 0.0
    farr[:, nx - 1] = 0.0

    result = n.fft.irfft2(farr)
    return trimarray(result, shape) if result.shape != shape else result
